"""
Main Knowledge Extractor class.

This works by recursively constructing structured prompt-completions where
a pseudo-YAML structure is requested, where the YAML
structure corresponds to a template class.

Describes in the SPIRES manuscript
TODO: add link
"""
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import pydantic
from linkml_runtime.linkml_model import ClassDefinition, SlotDefinition

from ontogpt.engines.knowledge_engine import (
    ANNOTATION_KEY_PROMPT,
    ANNOTATION_KEY_PROMPT_SKIP,
    EXAMPLE,
    FIELD,
    OBJECT,
    KnowledgeEngine,
    chunk_text,
)
from ontogpt.templates.core import ExtractionResult

this_path = Path(__file__).parent


RESPONSE_ATOM = Union[str, "ResponseAtom"]
RESPONSE_DICT = Dict[FIELD, Union[RESPONSE_ATOM, List[RESPONSE_ATOM]]]


@dataclass
class SPIRESEngine(KnowledgeEngine):
    """Knowledge extractor."""

    engine: str = "text-davinci-003"
    recurse: bool = True
    """If true, then complex non-named entity objects are always recursively parsed.
    If this is false AND the complex object is a pair, then token-based splitting is
    instead used.
    TODO: deprecate this, it's not clear that token-based splitting is better, due to
    the inability to control which tokens GPT will use"""

    sentences_per_window: Optional[int] = None
    """If set, this will split the text into chains of sentences,
    where this determines the maximum number of sentences per chain.
    The results are then merged together."""

    def extract_from_text(
        self, text: str, cls: ClassDefinition = None, object: OBJECT = None
    ) -> ExtractionResult:
        """
        Extract annotations from the given text.

        :param text:
        :param cls:
        :param object: optional stub object
        :return:
        """
        if self.sentences_per_window:
            chunks = chunk_text(text, self.sentences_per_window)
            extracted_object = None
            for chunk in chunks:
                raw_text = self._raw_extract(chunk, cls, object=object)
                logging.info(f"RAW TEXT: {raw_text}")
                next_object = self.parse_completion_payload(raw_text, cls, object=object)
                if extracted_object is None:
                    extracted_object = next_object
                else:
                    for k, v in next_object.items():
                        if isinstance(v, list):
                            extracted_object[k] += v
                        else:
                            if k not in extracted_object:
                                extracted_object[k] = v
                            else:
                                extracted_object[k] = v
        else:
            raw_text = self._raw_extract(text, cls, object=object)
            logging.info(f"RAW TEXT: {raw_text}")
            extracted_object = self.parse_completion_payload(raw_text, cls, object=object)
        return ExtractionResult(
            input_text=text,
            raw_completion_output=raw_text,
            prompt=self.last_prompt,
            extracted_object=extracted_object,
            named_entities=self.named_entities,
        )

    def _extract_from_text_to_dict(self, text: str, cls: ClassDefinition = None) -> RESPONSE_DICT:
        raw_text = self._raw_extract(text, cls)
        return self._parse_response_to_dict(raw_text, cls)

    def generalize(
        self, object: Union[pydantic.BaseModel, dict], examples: List[EXAMPLE]
    ) -> ExtractionResult:
        """
        Generalize the given examples.

        :param object:
        :param examples:
        :return:
        """
        cls = self.template_class
        sv = self.schemaview
        prompt = "example:\n"
        for example in examples:
            prompt += f"{self.serialize_object(example)}\n\n"
        prompt += "\n\n===\n\n"
        if isinstance(object, pydantic.BaseModel):
            object = object.dict()
        for k, v in object.items():
            if v:
                slot = sv.induced_slot(k, cls.name)
                prompt += f"{k}: {self._serialize_value(v, slot)}\n"
        logging.debug(f"PROMPT: {prompt}")
        payload = self.client.complete(prompt)
        prediction = self.parse_completion_payload(payload, object=object)
        return ExtractionResult(
            input_text=prompt,
            raw_completion_output=payload,
            # prompt=self.last_prompt,
            results=[prediction],
            named_entities=self.named_entities,
        )

    def map_terms(self, terms: List[str], ontology: str) -> Dict[str, List[str]]:
        """
        Map the given terms to the given ontology.

        EXPERIMENTAL

        currently GPT-3 does not do so well with this task.

        :param terms:
        :param ontology:
        :return:
        """
        # TODO: make a separate config
        examples = {
            "go": {
                "nucleui": "nucleus",
                "mitochondrial": "mitochondrion",
                "signaling": "signaling pathway",
                "cysteine biosynthesis": "cysteine biosynthetic process",
                "alcohol dehydrogenase": "alcohol dehydrogenase activity",
            },
            "uberon": {
                "feet": "pes",
                "forelimb, left": "left forelimb",
                "hippocampus": "Ammons horn",
            },
        }
        ontology = ontology.lower()
        if ontology in examples:
            example = examples[ontology]
        else:
            example = examples["uberon"]
        prompt = "Normalize the following semicolon separated\
            list of terms to the {ontology.upper()} ontology\n\n"
        prompt += "For example:\n\n"
        for k, v in example.items():
            prompt += f"{k}: {v}\n"
        prompt += "===\n\nTerms:"
        prompt += "; ".join(terms)
        prompt += "===\n\n"
        payload = self.client.complete(prompt)
        # outer parse
        best_results = []
        for sep in ["\n", "; "]:
            results = payload.split(sep)
            if len(results) > len(best_results):
                best_results = results

        def normalize(s: str) -> str:
            s = s.strip()
            s.replace("_", " ")
            return s.lower()

        mappings = {}
        for result in best_results:
            if ":" not in result:
                logging.error(f"Count not parse result: {result}")
                continue
            k, v = result.strip().split(":", 1)
            k = k.strip()
            v = v.strip()
            for t in terms:
                if normalize(t) == normalize(k):
                    mappings[t] = v
                    break
        for t in terms:
            if t not in mappings:
                logging.warning(f"Could not map term: {t}")
        return mappings

    def serialize_object(self, example: EXAMPLE, cls: ClassDefinition = None) -> str:
        if cls is None:
            cls = self.template_class
        if isinstance(example, str):
            return example
        if isinstance(example, pydantic.BaseModel):
            example = example.dict()
        lines = []
        sv = self.schemaview
        for k, v in example.items():
            if not v:
                continue
            slot = sv.induced_slot(k, cls.name)
            v_serialized = self._serialize_value(v, slot)
            lines.append(f"{k}: {v_serialized}")
        return "\n".join(lines)

    def _serialize_value(self, val: Any, slot: SlotDefinition) -> str:
        if val is None:
            return ""
        if isinstance(val, list):
            return "; ".join([self._serialize_value(v, slot) for v in val if v])
        if isinstance(val, dict):
            return " - ".join([self._serialize_value(v, slot) for v in val.values() if v])
        sv = self.schemaview
        if slot.range in sv.all_classes():
            if self.labelers:
                labelers = list(self.labelers)
            else:
                labelers = []
            labelers += self.get_annotators(sv.get_class(slot.range))
            if labelers:
                for labeler in labelers:
                    label = labeler.label(val)
                    if label:
                        return label
        return val

    def _raw_extract(self, text, cls: ClassDefinition = None, object: OBJECT = None) -> str:
        """
        Extract annotations from the given text.

        :param text:
        :return:
        """
        prompt = self.get_completion_prompt(cls, text, object=object)
        self.last_prompt = prompt
        payload = self.client.complete(prompt)
        return payload

    def get_completion_prompt(
        self, cls: ClassDefinition = None, text: str = None, object: OBJECT = None
    ) -> str:
        """Get the prompt for the given template."""
        if cls is None:
            cls = self.template_class
        if not text or ("\n" in text or len(text) > 60):
            prompt = (
                "From the text below, extract the following entities in the following format:\n\n"
            )
        else:
            prompt = "Split the following piece of text into fields in the following format:\n\n"
        for slot in self.schemaview.class_induced_slots(cls.name):
            if ANNOTATION_KEY_PROMPT_SKIP in slot.annotations:
                continue
            if ANNOTATION_KEY_PROMPT in slot.annotations:
                slot_prompt = slot.annotations[ANNOTATION_KEY_PROMPT].value
            elif slot.description:
                slot_prompt = slot.description
            else:
                if slot.multivalued:
                    slot_prompt = f"semicolon-separated list of {slot.name}s"
                else:
                    slot_prompt = f"the value for {slot.name}"
            if slot.range in self.schemaview.all_enums():
                enum_def = self.schemaview.get_enum(slot.range)
                pvs = [str(k) for k in enum_def.permissible_values.keys()]
                slot_prompt += f"Must be one of: {', '.join(pvs)}"
            prompt += f"{slot.name}: <{slot_prompt}>\n"
        # prompt += "Do not answer if you don't know\n\n"
        prompt = f"{prompt}\n\nText:\n{text}\n\n===\n\n"
        if object:
            if cls is None:
                cls = self.template_class
            if isinstance(object, pydantic.BaseModel):
                object = object.dict()
            for k, v in object.items():
                if v:
                    slot = self.schemaview.induced_slot(k, cls.name)
                    prompt += f"{k}: {self._serialize_value(v, slot)}\n"
        return prompt

    def _parse_response_to_dict(
        self, results: str, cls: ClassDefinition = None
    ) -> Optional[RESPONSE_DICT]:
        """
        Parse the pseudo-YAML response from OpenAI into a dictionary object.

        E.g.

            foo: a; b; c

        becomes

            {"foo": ["a", "b", "c"]}

        :param results:
        :return:
        """
        lines = results.splitlines()
        ann = {}
        promptable_slots = self.promptable_slots(cls)
        for line in lines:
            line = line.strip()
            if not line:
                continue
            if ":" not in line:
                if len(promptable_slots) == 1:
                    slot = promptable_slots[0]
                    logging.warning(
                        f"Coercing to YAML-like with key {slot.name}: Original line: {line}"
                    )
                    line = f"{slot.name}: {line}"
                else:
                    logging.error(f"Line '{line}' does not contain a colon; ignoring")
                    return
            r = self._parse_line_to_dict(line, cls)
            if r is not None:
                field, val = r
                ann[field] = val
        return ann

    def _parse_line_to_dict(
        self, line: str, cls: ClassDefinition = None
    ) -> Optional[Tuple[FIELD, RESPONSE_ATOM]]:
        if cls is None:
            cls = self.template_class
        sv = self.schemaview
        # each line is a key-value pair
        logging.info(f"PARSING LINE: {line}")
        field, val = line.split(":", 1)
        # Field nornalization:
        # The LLML may mutate the output format somewhat,
        # randomly pluralizing or replacing spaces with underscores
        field = field.lower().replace(" ", "_")
        cls_slots = sv.class_slots(cls.name)
        slot = None
        if field in cls_slots:
            slot = sv.induced_slot(field, cls.name)
        else:
            if field.endswith("s"):
                field = field[:-1]
            if field in cls_slots:
                slot = sv.induced_slot(field, cls.name)
        if not slot:
            logging.error(f"Cannot find slot for {field} in {line}")
            # raise ValueError(f"Cannot find slot for {field} in {line}")
            return
        if not val:
            msg = f"Empty value in key-value line: {line}"
            if slot.required:
                raise ValueError(msg)
            if slot.recommended:
                logging.warning(msg)
            return
        inlined = slot.inlined
        slot_range = sv.get_class(slot.range)
        if not inlined:
            if slot.range in sv.all_classes():
                inlined = sv.get_identifier_slot(slot_range.name) is None
        val = val.strip()
        if slot.multivalued:
            vals = [v.strip() for v in val.split(";")]
        else:
            vals = [val]
        vals = [val for val in vals if val]
        logging.debug(f"SLOT: {slot.name} INL: {inlined} VALS: {vals}")
        if inlined:
            transformed = False
            slots_of_range = sv.class_slots(slot_range.name)
            if self.recurse or len(slots_of_range) > 2:
                vals = [self._extract_from_text_to_dict(v, slot_range) for v in vals]
            else:
                for sep in [" - ", ":", "/", "*", "-"]:
                    if all([sep in v for v in vals]):
                        vals = [dict(zip(slots_of_range, v.split(sep, 1))) for v in vals]
                        for v in vals:
                            for k in v.keys():
                                v[k] = v[k].strip()
                        transformed = True
                        break
                if not transformed:
                    logging.warning(f"Did not find separator in {vals} for line {line}")
                    return
        # transform back from list to single value if not multivalued
        if slot.multivalued:
            final_val = vals
        else:
            if len(vals) != 1:
                logging.error(f"Expected 1 value for {slot.name} in '{line}' but got {vals}")
            final_val = vals[0]
        return field, final_val

    def parse_completion_payload(
        self, results: str, cls: ClassDefinition = None, object: dict = None
    ) -> pydantic.BaseModel:
        """
        Parse the completion payload into a pydantic class.

        :param results:
        :param cls:
        :param object: stub object
        :return:
        """
        raw = self._parse_response_to_dict(results, cls)
        logging.debug(f"RAW: {raw}")
        if object:
            raw = {**object, **raw}
        return self.ground_annotation_object(raw, cls)

    def ground_annotation_object(
        self, ann: RESPONSE_DICT, cls: ClassDefinition = None
    ) -> Optional[pydantic.BaseModel]:
        """Ground the direct parse of the OpenAI payload.

        The raw openAI payload is a YAML-like string, which is parsed to
        a response dictionary.

        This dictionary is then grounded, using this method

        :param ann: Raw annotation object
        :param cls: schema class the ground object should instantiate
        :return: Grounded annotation object
        """
        logging.debug(f"Grounding annotation object {ann}")
        if cls is None:
            cls = self.template_class
        sv = self.schemaview
        new_ann = {}
        if ann is None:
            logging.error(f"Cannot ground None annotation, cls={cls.name}")
            return
        for field, vals in ann.items():
            if isinstance(vals, list):
                multivalued = True
            else:
                multivalued = False
                vals = [vals]
            slot = sv.induced_slot(field, cls.name)
            rng_cls = sv.get_class(slot.range)
            enum_def = None
            if slot.range:
                if slot.range in self.schemaview.all_enums():
                    enum_def = self.schemaview.get_enum(slot.range)
            new_ann[field] = []
            for val in vals:
                if not val:
                    continue
                if isinstance(val, tuple):
                    # special case for pairs
                    sub_slots = sv.class_induced_slots(rng_cls.name)
                    obj = {}
                    for i in range(0, len(val)):
                        sub_slot = sub_slots[i]
                        sub_rng = sv.get_class(sub_slot.range)
                        if not sub_rng:
                            logging.error(f"Cannot find range for {sub_slot.name}")
                        result = self.normalize_named_entity(val[i], sub_slot.range)
                        obj[sub_slot.name] = result
                elif isinstance(val, dict):
                    # recurse
                    obj = self.ground_annotation_object(val, rng_cls)
                else:
                    obj = self.normalize_named_entity(val, slot.range)
                if enum_def:
                    found = False
                    logging.info(f"Looking for {obj} in {enum_def.name}")
                    for k, _pv in enum_def.permissible_values.items():
                        if obj.lower() == k.lower():
                            obj = k
                            found = True
                            break
                    if not found:
                        logging.info(f"Cannot find enum value for {obj} in {enum_def.name}")
                        obj = None
                if multivalued:
                    new_ann[field].append(obj)
                else:
                    new_ann[field] = obj
        logging.debug(f"Creating object from dict {new_ann}")
        logging.info(new_ann)
        py_cls = self.template_module.__dict__[cls.name]
        return py_cls(**new_ann)



"""Eval enrichment template."""
import datetime
import gzip
import logging
import random
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Tuple

import tiktoken
import yaml
from cachier import cachier
from linkml_runtime.dumpers import json_dumper
from oaklib import get_adapter, get_implementation_from_shorthand
from oaklib.datamodels.association import Association
from oaklib.datamodels.vocabulary import EQUIVALENT_CLASS, IS_A, PART_OF
from oaklib.interfaces.class_enrichment_calculation_interface import (
    ClassEnrichmentCalculationInterface,
)
from oaklib.interfaces.obograph_interface import OboGraphInterface
from oaklib.parsers.association_parser_factory import get_association_parser
from pydantic import BaseModel
from tiktoken import Encoding

from ontogpt.engines import create_engine
from ontogpt.engines.enrichment import ENTITY_ID, EnrichmentEngine, EnrichmentPayload
from ontogpt.engines.knowledge_engine import MODEL_GPT_3_5_TURBO, MODEL_NAME, MODEL_TEXT_DAVINCI_003, MODEL_GPT_4
from ontogpt.evaluation.evaluation_engine import EvaluationEngine
from ontogpt.templates.class_enrichment import ClassEnrichmentResult
from ontogpt.utils.gene_set_utils import SYMBOL, GeneSet, drop_genes_from_gene_set, gene_info

THIS_DIR = Path(__file__).parent
DATABASE_DIR = Path(__file__).parent / "database"

ONTOLOGICAL_SYNOPSIS = "ontological_synopsis"
NARRATIVE_SYNOPSIS = "narrative_synopsis"
NO_SYNOPSIS = "no_synopsis"
STANDARD = "standard"
STANDARD_NO_ONTOLOGY = "standard_no_ontology"
RANDOM = "random"
RANK_BASED = "rank_based"
CLOSURE = "closure"

ENRICHMENT_MODELS = [MODEL_GPT_4, MODEL_GPT_3_5_TURBO, MODEL_TEXT_DAVINCI_003]


logger = logging.getLogger(__name__)


class Overlap(BaseModel):
    """Overlap between two sets."""

    jaccard: Optional[float] = None
    common: Optional[List[str]] = None
    overlap_score: Optional[int] = None
    left_jaccard: float = None
    right_jaccard: float = None
    summary_jaccard: float = None


class GeneSetComparison(BaseModel):
    name: str
    gene_symbols: List[str]
    gene_ids: List[str] = None
    model: str = None
    payloads: Dict[str, EnrichmentPayload] = None
    overlaps: Dict[Tuple[str, str], Overlap] = None
    number_of_genes_swapped_out: int = None


@cachier(stale_after=datetime.timedelta(days=3))
def get_symbol_to_gene_id_map() -> Dict[SYMBOL, ENTITY_ID]:
    hgnc = get_adapter("sqlite:obo:hgnc")
    label2id = {s: id.upper() for id, s in hgnc.labels(hgnc.entities())}
    return label2id


@dataclass
class EvalEnrichment(EvaluationEngine):
    tokenizer_encoding: Encoding = None

    loaded: bool = False

    model: str = None

    engines: Dict[MODEL_NAME, EnrichmentEngine] = field(default_factory=dict)

    def __post_init__(self):
        ontology = get_implementation_from_shorthand("sqlite:obo:go")
        if not isinstance(ontology, OboGraphInterface):
            raise TypeError
        self.ontology: ClassEnrichmentCalculationInterface = ontology
        self.engine: EnrichmentEngine = create_engine(None, EnrichmentEngine, model=self.model)
        for model in ENRICHMENT_MODELS:
            self.engines[model] = create_engine(None, EnrichmentEngine, model=model)
            self.engines[model].add_resolver("sqlite:obo:hgnc")
        self.engine.add_resolver("sqlite:obo:hgnc")
        self.tokenizer_encoding = tiktoken.encoding_for_model(self.engine.model)

    def evaluate_methods_on_gene_set(
        self, gene_set: GeneSet, max_size=999, n=4, randomize_gene_descriptions=False, **kwargs
    ) -> List[GeneSetComparison]:
        """
        Perform evaluation of different methods on a gene set.

        Different permutations of the gene set are created, dropping out some members, and
        inserting random genes.

        :param gene_set: Gene set to evaluate
        :param max_size: Maximum size of gene set, above this the gene set will be truncated
        :param n: Number of dropouts to perform
        :param randomize_gene_descriptions: use random gene descriptions
        """
        if n < 1:
            raise ValueError(f"n must be greater than 0: {n}")
        if n > 5:
            raise ValueError(f"n must be less than 5: {n}")
        name = gene_set.name
        comparisons = []
        for i in range(0, n):
            perturbed_gene_set = drop_genes_from_gene_set(gene_set, i * 10)
            perturbed_gene_set.name = f"{name}-{i}"
            comp = self.compare_analysis(
                perturbed_gene_set,
            )
            logger.debug(comp)
            logger.info(yaml.dump(comp.dict(), sort_keys=False))
            comparisons.append(comp)
        return comparisons

    def compare_analysis(self, gene_set: GeneSet, **kwargs) -> GeneSetComparison:
        """Compare OntoGPT enrichment vs standard."""
        payloads = {}
        logger.info(f"Gene symbols: {gene_set.gene_symbols}")
        for model in ENRICHMENT_MODELS:
            engine = self.engines[model]
            for method in [NO_SYNOPSIS, ONTOLOGICAL_SYNOPSIS, NARRATIVE_SYNOPSIS]:
                if method == ONTOLOGICAL_SYNOPSIS:
                    args = dict(ontological_synopsis=True)
                elif method == NARRATIVE_SYNOPSIS:
                    args = dict(ontological_synopsis=False)
                elif method == NO_SYNOPSIS:
                    args = dict(annotations=False)
                else:
                    raise AssertionError(f"Unknown method: {method}")
                for prompt_variant, end_marker in [("v1", "==="), ("v2", "###")]:
                    engine.end_marker = end_marker
                    payload = engine.summarize(gene_set, normalize=True, **args)
                    payload.method = method
                    payload.prompt_variant = prompt_variant
                    model_method = f"{model}.{method}.{prompt_variant}"
                    payloads[model_method] = payload
        payloads[STANDARD] = self.standard_enrichment(gene_set)
        payloads[STANDARD_NO_ONTOLOGY] = self.standard_enrichment(gene_set, use_ontology=False)
        payloads[RANDOM] = self.random_enrichment(gene_set)
        payloads[RANK_BASED] = self.null_enrichment(gene_set)
        payloads[CLOSURE] = self.gene_term_closure(gene_set)
        for k in [STANDARD, STANDARD_NO_ONTOLOGY, RANDOM, RANK_BASED]:
            payloads[k].method = k
        comp = GeneSetComparison(
            name=gene_set.name,
            gene_symbols=gene_set.gene_symbols,
            gene_ids=gene_set.gene_ids,
            number_of_genes_swapped_out=gene_set.number_of_genes_dropped,
            payloads=payloads,
            **kwargs,
        )
        return comp

    def random_gene_symbol(self) -> ENTITY_ID:
        """Get a random gene."""
        assocs = list(self.ontology.associations())
        logger.debug(f"Got {len(assocs)} associations")
        ann = random.SystemRandom().choice(assocs)
        info = gene_info(ann.subject)
        return info[0]

    def standard_enrichment(self, gene_set: GeneSet, use_ontology=True) -> EnrichmentPayload:
        """Perform standard gene set enrichment.

        :param gene_set: GeneSet object, with gene_ids populated
        """
        gene_ids = gene_set.gene_ids
        if use_ontology:
            predicates = [IS_A, PART_OF]
        else:
            predicates = [EQUIVALENT_CLASS]  # TODO
        results = self.ontology.enriched_classes(
            gene_ids, autolabel=True, object_closure_predicates=predicates
        )
        # awkward: convert from oaklib dataclasses to pydantic
        results = [ClassEnrichmentResult(**json_dumper.to_dict(result)) for result in results]
        payload = EnrichmentPayload(enrichment_results=results)
        for result in results:
            if result.p_value_adjusted > 0.05:
                continue
            payload.term_strings.append(result.class_label)
            payload.term_ids.append(result.class_id)
        return payload

    def gene_term_closure(self, gene_set: GeneSet) -> EnrichmentPayload:
        gene_ids = gene_set.gene_ids
        term_ids = {a.object for a in self.ontology.associations(subjects=gene_ids)}
        ancs = list(self.ontology.ancestors(list(term_ids), [IS_A, PART_OF]))
        payload = EnrichmentPayload()
        for anc in ancs:
            payload.term_ids.append(anc)
            payload.term_strings.append(self.ontology.label(anc))
        return payload

    def random_enrichment(self, gene_set: GeneSet = None, n: int = None) -> EnrichmentPayload:
        """Randomized enrichment results."""
        if n is None:
            # by default, return a number of terms proportional to the number of genes
            if gene_set:
                n = len(gene_set.gene_symbols)
            else:
                n = 20
        anns = list(self.ontology.associations())
        random.shuffle(anns)
        payload = EnrichmentPayload()
        term_ids = list({ann.object for ann in anns[:n]})
        payload.term_ids = term_ids
        payload.term_strings = [self.ontology.label(id) for id in term_ids]
        return payload

    def null_enrichment(self, gene_set: GeneSet, n: int = None) -> EnrichmentPayload:
        """Psuedo-enrichment, returning all top ranking direct terms, no ontology rollup."""
        gene_symbols = gene_set.gene_symbols
        if n is None:
            # by default, return a number of terms proportional to the number of genes
            n = len(gene_symbols)
        anns = list(self.ontology.associations())
        counts = defaultdict(int)
        for ann in anns:
            counts[ann.object] += 1
        sorted_counts = sorted(counts.items(), key=lambda x: x[1], reverse=True)
        payload = EnrichmentPayload()
        term_ids = list({term for term, _ in sorted_counts[:n]})
        payload.term_ids = term_ids
        payload.term_strings = [self.ontology.label(id) for id in term_ids]
        return payload

    def create_gene_set_from_term(self, term: ENTITY_ID, name: str = None) -> GeneSet:
        """Create a gene set from a list of terms."""
        assocs = self.ontology.associations(
            objects=[term], object_closure_predicates=[IS_A, PART_OF]
        )
        gene_ids = list(set([str(assoc.subject) for assoc in assocs]))
        hgnc = get_adapter("sqlite:obo:hgnc")
        gene_symbols = [hgnc.label(id) for id in gene_ids]
        gene_symbols = [str(sym) for sym in gene_symbols if sym is not None]
        if name is None:
            name = f"term-{term}"
        return GeneSet(name=name, gene_symbols=gene_symbols, gene_ids=gene_ids)

    def compare_payloads(self, a: EnrichmentPayload, b: EnrichmentPayload) -> Overlap:
        """Compare two payloads."""
        a_ids = set(a.term_ids)
        b_ids = set(b.term_ids)
        overlap = a_ids.intersection(b_ids)
        len_overlap = len(overlap)
        len_union = len(a_ids) + len(b_ids) - len_overlap
        ov = Overlap(
            common=list(overlap),
            overlap_score=len_overlap,
            left_jaccard=len_overlap / len(a_ids) if len(a_ids) else 0,
            right_jaccard=len_overlap / len(b_ids) if len(b_ids) else 0,
            jaccard=len_overlap / len_union if len_union else 0,
        )
        if a.summary and b.summary:
            a_toks = set(self.tokenizer_encoding.encode(a.summary))
            b_toks = set(self.tokenizer_encoding.encode(b.summary))
            overlap_toks = a_toks.intersection(b_toks)
            len_overlap_toks = len(overlap_toks)
            union_toks = a_toks.union(b_toks)
            len_union_toks = len(union_toks)
            ov.summary_jaccard = len_overlap_toks / len_union_toks if len_union_toks else 0
        return ov

    def get_annotation_tuples(self, path=None) -> Tuple[str, str]:
        """Load."""
        if path is None:
            DATABASE_DIR / "gene2go.tsv.gz"
        with gzip.open(str(path), "rt") as f:
            for line in f.readlines():
                line = line.strip("\n")
                yield tuple(line.split("\t"))

    def get_mapped_annotations(self, path=None) -> Iterator[Tuple[str, str]]:
        """Load."""
        tupls = list(self.get_annotation_tuples(path))
        symbols = set([sym for sym, _ in tupls])
        m = map_hgnc_symbols(tuple(symbols))
        for sym, _ in tupls:
            if sym in m:
                yield sym, m[sym]
                continue

    def load_annotations(self, path=None, format=None) -> None:
        """
        Load annotations from a two-column TSV file or a GAF.

        :param path:
        :return:
        """
        if self.loaded:
            return
        assocs = []
        if not format:
            if path.endswith(".tsv.gz"):
                format = "tsv.gz"
            elif path.endswith(".gaf"):
                format = "gaf"
            else:
                raise ValueError(f"Unknown format: {format}")
        if format == "tsv.gz":
            # We have a slightly awkward special case for HGNC as the GO GAFs
            # use UniProtKB IDs
            m = get_symbol_to_gene_id_map()
            for sym, term_id in self.get_annotation_tuples(path):
                if sym in m:
                    gene_id = m[sym]
                    assocs.append(Association(subject=gene_id, object=term_id))
        elif format == "gaf":
            association_parser = get_association_parser("gaf")
            with open(path) as file:
                assocs = list(association_parser.parse(file))
        else:
            raise ValueError(f"Unknown format: {format}")
        self.ontology.add_associations(assocs)
        self.loaded = True



"""Pubmed Client."""
import logging
from dataclasses import dataclass, field
from typing import Iterator, List

import inflection
from eutils import Client
from eutils._internal.xmlfacades.pubmedarticle import PubmedArticle

PMID = str
TITLE_WEIGHT = 5
MAX_PMIDS = 50


def _normalize(s: str) -> str:
    return inflection.singularize(s).lower()


def _score_paper(paper: PubmedArticle, keywords: List[str]) -> int:
    title_score = _score_text(paper.title, keywords)
    abstract_score = _score_text(paper.abstract, keywords)
    logging.info(f"Scored {paper.pmid} {paper.title} with TS={title_score} AS={abstract_score} ")
    return title_score * TITLE_WEIGHT + abstract_score


def _score_text(text: str, keywords: List[str]) -> int:
    text = text.lower()
    if not text:
        return -100
    score = 0
    for kw in keywords:
        if kw in text:
            score += 1
    return score


@dataclass
class PubmedClient:
    """A client for the Pubmed API.

    This class is a wrapper around the Entrez API.
    """

    entrez_client: Client = field(default_factory=lambda: Client())
    max_text_length = 3000

    def text(self, id: PMID, autoformat=True) -> str:
        """Get the text of a paper from its PMID.

        :param id:
        :param autoformat: if True include title and abstract concatenated
        :return:
        """
        ec = self.entrez_client
        id = id.replace("PMID:", "")
        paset = ec.efetch(db="pubmed", id=id)
        for pa in paset:
            if autoformat:
                txt = f"Title: {pa.title}\nAbstract: {pa.abstract}\nKeywords: {'; '.join(pa.mesh_headings)}"  # noqa
            else:
                txt = pa.full_text
        if len(txt) > self.max_text_length:
            logging.warning(f"Truncating text: {txt[:self.max_text_length]}...")
            txt = txt[0 : self.max_text_length]
        return txt

    def search(self, term: str, keywords: List[str] = None) -> Iterator[PMID]:
        """Get the text of a paper from its PMID.

        :param term:
        :param keywords:
        :return:
        """
        print("Getting client")
        ec = self.entrez_client
        if keywords:
            keywords = [_normalize(kw) for kw in keywords]
            term = f"({term}) AND ({' OR '.join(keywords)})"
        logging.info(f"Searching for {term}...")
        esr = ec.esearch(db="pubmed", term=term)
        logging.info(f"Found {esr.count} papers for {term}.")
        paset = ec.efetch(db="pubmed", id=esr.ids[0:MAX_PMIDS])
        keywords = keywords or []
        keywords = [_normalize(kw) for kw in keywords]
        scored_papers = [(_score_paper(paper, keywords), paper) for paper in paset]
        scored_papers.sort(key=lambda x: x[0], reverse=True)
        for score, paper in scored_papers:
            logging.debug(f"Yielding {paper.pmid} {paper.title} with score {score} ")
            yield f"PMID:{paper.pmid}"



"""Exporter."""
from dataclasses import dataclass
from pathlib import Path
from typing import TextIO, Union

from ontogpt.templates.core import ExtractionResult


def is_curie(s: str) -> bool:
    return ":" in s and " " not in s


@dataclass
class Exporter:
    def export(self, extraction_output: ExtractionResult, output: Union[str, Path, TextIO]):
        raise NotImplementedError



"""YAML Wrapper."""
from typing import Any

import pydantic
import yaml


def eliminate_empty(obj: Any, preserve=False) -> Any:
    """Eliminate empty lists and dicts from an object."""
    if isinstance(obj, list):
        return [eliminate_empty(x, preserve) for x in obj if x or preserve]
    elif isinstance(obj, dict):
        return {k: eliminate_empty(v, preserve) for k, v in obj.items() if v or preserve}
    elif isinstance(obj, pydantic.BaseModel):
        return eliminate_empty(obj.dict(), preserve)
    elif isinstance(obj, tuple):
        return [eliminate_empty(x, preserve) for x in obj]
    elif isinstance(obj, str):
        return str(obj)
    else:
        return obj


def dump_minimal_yaml(obj: Any, minimize=True) -> str:
    """Dump a YAML string, but eliminating Nones and empty lists and dicts."""
    return yaml.dump(eliminate_empty(obj, not minimize), sort_keys=False)



"""Main Knowledge Extractor class."""
import importlib
import logging
import re
from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path
from types import ModuleType
from typing import Dict, Iterator, List, Optional, TextIO, Union
from urllib.parse import quote

import inflection
import openai
import pydantic
import tiktoken
import yaml
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import ClassDefinition, ElementName, SlotDefinition
from oaklib import BasicOntologyInterface, get_adapter
from oaklib.datamodels.text_annotator import TextAnnotationConfiguration
from oaklib.implementations import OntoPortalImplementationBase
from oaklib.interfaces import MappingProviderInterface, TextAnnotatorInterface
from oaklib.utilities.apikey_manager import get_apikey_value
from oaklib.utilities.subsets.value_set_expander import ValueSetExpander

from ontogpt.clients import OpenAIClient
from ontogpt.templates.core import ExtractionResult, NamedEntity

this_path = Path(__file__).parent
logger = logging.getLogger(__name__)


OBJECT = Union[str, pydantic.BaseModel, dict]
EXAMPLE = OBJECT
FIELD = str
TEMPLATE_NAME = str
MODEL_NAME = str

# annotation metamodel
ANNOTATION_KEY_PROMPT = "prompt"
ANNOTATION_KEY_PROMPT_SKIP = "prompt.skip"
ANNOTATION_KEY_ANNOTATORS = "annotators"
ANNOTATION_KEY_RECURSE = "ner.recurse"
ANNOTATION_KEY_EXAMPLES = "prompt.examples"

MODEL_GPT_3_5_TURBO = "gpt-3.5-turbo"
MODEL_TEXT_DAVINCI_003 = "text-davinci-003"
MODEL_GPT_4 = "gpt-4"
MODELS = [MODEL_GPT_3_5_TURBO, MODEL_TEXT_DAVINCI_003, MODEL_GPT_4]

DEFAULT_MODEL = MODEL_GPT_3_5_TURBO

# TODO: introspect
DATAMODELS = [
    "treatment.DiseaseTreatmentSummary",
    "gocam.GoCamAnnotations",
    "bioloigical_process.BiologicalProcess",
    "environmental_sample.Study",
    "mendelian_disease.MendelianDisease",
    "reaction.Reaction",
    "recipe.Recipe",
]


def chunk_text(text: str, window_size=3) -> Iterator[str]:
    """Chunk text into windows of sentences."""
    sentences = re.split(r"[.?!]\s+", text)
    for right_index in range(1, len(sentences)):
        left_index = max(0, right_index - window_size)
        yield ". ".join(sentences[left_index:right_index])


@dataclass
class KnowledgeEngine(ABC):
    """
    Abstract base class for all knowledge engines.

    A Knowledge Engine is able to extract knowledge from text, utilizing
    knowledge sources plus LLMs
    """

    template: TEMPLATE_NAME = None
    """LinkML Template to use for this engine.
    Must be of the form <module_name>.<ClassName>"""

    template_class: ClassDefinition = None
    """LinkML Class for the template.
    This is derived from the template and does not need to be set manually."""

    template_pyclass = None
    """Python class for the template.
    This is derived from the template and does not need to be set manually."""

    template_module: ModuleType = None
    """Python module for the template.
    This is derived from the template and does not need to be set manually."""

    schemaview: SchemaView = None
    """LinkML SchemaView over the template.
    This is derived from the template and does not need to be set manually."""

    api_key: str = None
    """OpenAI API key."""

    model: MODEL_NAME = None
    """OpenAI Model. This should be overridden in subclasses"""

    # annotator: TextAnnotatorInterface = None
    # """Default annotator. TODO: deprecate?"""

    annotators: Dict[str, List[TextAnnotatorInterface]] = None
    """Annotators for each class.
    An annotator will ground/map labels to CURIEs.
    These override the annotators annotated in the template
    """

    skip_annotators: Optional[List[TextAnnotatorInterface]] = None
    """Annotators to skip.
    This overrides any specified in the schema"""

    mappers: List[BasicOntologyInterface] = None
    """List of concept mappers, to assist in grounding to desired ID prefix"""

    labelers: List[BasicOntologyInterface] = None
    """Labelers that map CURIEs to labels"""

    client: OpenAIClient = None
    """All calls to LLMs are delegated through this client"""

    dictionary: Dict[str, str] = field(default_factory=dict)
    """Local dictionary of strings/labels to IDs"""

    value_set_expansions: Dict[str, List[str]] = field(default_factory=dict)

    min_grounding_text_overlap = 0.66
    """Min proportion of overlap in characters between text and grounding. TODO: use tokenization"""

    named_entities: List[NamedEntity] = field(default_factory=list)
    """Cache of all named entities"""

    auto_prefix: str = None
    """If set then non-normalized named entities will be mapped to this prefix"""

    last_text: str = None
    """Cache of last text."""

    last_prompt: str = None
    """Cache of last prompt used."""

    encoding = None

    def __post_init__(self):
        if self.template:
            self.template_class = self._get_template_class(self.template)
        if self.template_class:
            logging.info(f"Using template {self.template_class.name}")
        if not self.model:
            self.model = DEFAULT_MODEL
        self.client = OpenAIClient(model=self.model)
        logging.info("Setting up OpenAI client API Key")
        self.api_key = self._get_openai_api_key()
        openai.api_key = self.api_key
        if self.mappers is None:
            logging.info("Using mappers (currently hardcoded)")
            self.mappers = [get_adapter("translator:")]
        self.encoding = tiktoken.encoding_for_model(self.client.model)

    def extract_from_text(
        self, text: str, cls: ClassDefinition = None, object: OBJECT = None
    ) -> ExtractionResult:
        raise NotImplementedError

    def extract_from_file(self, file: Union[str, Path, TextIO]) -> pydantic.BaseModel:
        """
        Extract annotations from the given text.

        :param file:
        :return:
        """
        if isinstance(file, str):
            file = Path(file)
        if isinstance(file, Path):
            with file.open() as f:
                text = f.read()
        else:
            text = file.read()
        self.last_text = text
        r = self.extract_from_text(text)
        r.input_id = str(file)
        return r

    def load_dictionary(self, path: Union[str, Path, list]):
        if not isinstance(path, list):
            logger.info(f"Loading dictionary from {path}")
            with open(str(path)) as f:
                return self.load_dictionary(yaml.safe_load(f))
        if self.dictionary is None:
            self.dictionary = {}
        entries = [(entry["synonym"].lower(), entry["id"]) for entry in path]
        entries = sorted(entries, key=lambda x: len(x[0]), reverse=True)
        for syn, id in entries:
            if syn in self.dictionary and self.dictionary[syn] != id:
                logger.warning(f"Duplicate synonym: {syn} => {id}, {self.dictionary[syn]}")
            self.dictionary[syn] = id
        logger.info(f"Loaded {len(self.dictionary)}")

    # @abstractmethod
    def synthesize(self, cls: ClassDefinition = None, object: OBJECT = None) -> ExtractionResult:
        raise NotImplementedError

    def generalize(
        self, object: Union[pydantic.BaseModel, dict], examples: List[EXAMPLE]
    ) -> ExtractionResult:
        raise NotImplementedError

    def map_terms(self, terms: List[str], ontology: str) -> Dict[str, List[str]]:
        raise NotImplementedError

    def _get_template_class(self, template: TEMPLATE_NAME) -> ClassDefinition:
        """
        Get the LinkML class for a template.

        :param template: template name of the form module.ClassName
        :return: LinkML class definition
        """
        logger.info(f"Loading schema for {template}")
        if "." in template:
            module_name, class_name = template.split(".", 1)
        else:
            module_name = template
            class_name = None
        templates_path = this_path.parent / "templates"
        path_to_template = str(templates_path / f"{module_name}.yaml")
        sv = SchemaView(path_to_template)
        if class_name is None:
            roots = [c.name for c in sv.all_classes().values() if c.tree_root]
            if len(roots) != 1:
                raise ValueError(f"Template {template} does not have singular root: {roots}")
            class_name = roots[0]
        mod = importlib.import_module(f"ontogpt.templates.{module_name}")
        self.template_module = mod
        self.template_pyclass = mod.__dict__[class_name]
        self.schemaview = sv
        logger.info(f"Getting class for template {template}")
        cls = None
        for c in sv.all_classes().values():
            if c.name == class_name:
                cls = c
                break
        if not cls:
            raise ValueError(f"Template {template} not found")
        return cls

    def _get_openai_api_key(self):
        """Get the OpenAI API key from the environment."""
        # return os.environ.get("OPENAI_API_KEY")
        return get_apikey_value("openai")

    def get_annotators(self, cls: ClassDefinition = None) -> List[BasicOntologyInterface]:
        """
        Get the annotators/labelers for a class.

        The annotators are returned in order of precedence

        Annotators are used to *ground* labels as CURIEs.
        Annotators may also do double-duty as labelers (i.e. map CURIEs to labels)

        These are specified by linkml annotations within the template/schema;
        if the engine has a set of annotators specified these take precedence.

        :param cls: schema class
        :return: list of annotations
        """
        if self.annotators and cls.name in self.annotators:
            annotators = self.annotators[cls.name]
        else:
            if ANNOTATION_KEY_ANNOTATORS not in cls.annotations:
                logger.error(f"No annotators for {cls.name}")
                return []
            annotators = cls.annotations[ANNOTATION_KEY_ANNOTATORS].value.split(", ")
        logger.info(f" Annotators: {annotators} [will skip: {self.skip_annotators}]")
        annotators = []
        for annotator in annotators:
            if isinstance(annotator, str):
                logger.info(f"Loading annotator {annotator}")
                if self.skip_annotators and annotator in self.skip_annotators:
                    logger.info(f"Skipping annotator {annotator}")
                    continue
                if annotator not in self.annotators:
                    self.annotators[annotator] = get_adapter(annotator)
                annotators.append(self.annotators[annotator])
            elif isinstance(annotator, BasicOntologyInterface):
                annotators.append(annotator)
            else:
                raise ValueError(f"Unknown annotator type {annotator}")
        return annotators

    def promptable_slots(self, cls: Optional[ClassDefinition] = None) -> List[SlotDefinition]:
        """
        List of all slots that are not skipped for purposes of prompting.

        Examples of slots that are skipped are:

        - identifier fields
        - the source text used in extraction
        - other metadata that is outside what we might want to predict

        :param cls:
        :return:
        """
        if cls is None:
            cls = self.template_class
        sv = self.schemaview
        return [s for s in sv.class_induced_slots(cls.name) if not self.slot_is_skipped(s)]

    def slot_is_skipped(self, slot: SlotDefinition) -> bool:
        sv = self.schemaview
        if ANNOTATION_KEY_PROMPT_SKIP in slot.annotations:
            return True

    def normalize_named_entity(self, text: str, range: ElementName) -> str:
        """
        Grounds and normalizes to preferred ID prefixes.

        if the entity cannot be grounded and normalized, the original text is returned.

        :param text:
        :param cls:
        :return:
        """
        sv = self.schemaview
        cls = sv.get_class(range)
        if cls is None:
            return text
        if ANNOTATION_KEY_EXAMPLES in cls.annotations:
            examples = cls.annotations[ANNOTATION_KEY_EXAMPLES].value.split(", ")
            examples = [x.lower() for x in examples]
            logger.debug(f"Will exclude if in list of examples: {examples}")
            if text.lower() in examples:
                logger.warning(f"Likely a hallucination as it is the example set: {text}")
                return f"LIKELY HALLUCINATION: {text}"
        for obj_id in self.groundings(text, cls):
            logger.info(f"Grounding {text} to {obj_id}; next step is to normalize")
            for normalized_id in self.normalize_identifier(obj_id, cls):
                if not any(e for e in self.named_entities if e.id == normalized_id):
                    self.named_entities.append(NamedEntity(id=normalized_id, label=text))
                logger.info(f"Normalized {text} with {obj_id} to {normalized_id}")
                return normalized_id
        logger.info(f"Could not ground and normalize {text} to {cls.name}")
        if self.auto_prefix:
            obj_id = f"{self.auto_prefix}:{quote(text)}"
            if not any(e for e in self.named_entities if e.id == obj_id):
                self.named_entities.append(NamedEntity(id=obj_id, label=text))
        else:
            obj_id = text
        if ANNOTATION_KEY_RECURSE in cls.annotations:
            logger.info(f"Using recursive strategy to parse: {text} to {cls.name}")
            obj = self.extract_from_text(text, cls).extracted_object
            if obj:
                if self.named_entities is None:
                    self.named_entities = []
                try:
                    obj.id = obj_id
                except ValueError as e:
                    logger.error(f"No id for {obj} {e}")
                self.named_entities.append(obj)
        return obj_id

    def is_valid_identifier(self, input_id: str, cls: ClassDefinition) -> bool:
        sv = self.schemaview
        if cls.id_prefixes:
            if ":" not in input_id:
                return False
            prefix, rest = input_id.split(":", 1)
            if prefix not in cls.id_prefixes:
                logger.debug(f"ID {input_id} not in prefixes {cls.id_prefixes}")
                return False
        id_slot = sv.get_identifier_slot(cls.name)
        if id_slot and id_slot.pattern:
            id_regex = re.compile(id_slot.pattern)
            m = re.match(id_regex, input_id)
            if not m:
                logger.debug(f"ID {input_id} does not match pattern {id_slot.pattern}")
                return False
        if id_slot and id_slot.values_from:
            vse = ValueSetExpander()
            is_found = False
            for e in id_slot.values_from:
                if e not in self.value_set_expansions:
                    # expanding value set for first time
                    range_enum = sv.get_enum(e)
                    pvs = vse.expand_value_set(range_enum, sv.schema)
                    valid_ids = [pv.text for pv in pvs]
                    self.value_set_expansions[e] = valid_ids
                    logger.info(f"Expanded {e} to {len(valid_ids)} IDs")
                if input_id in self.value_set_expansions[e]:
                    is_found = True
                    logger.info(f"ID {input_id} found in value set {e}")
                    break
            if not is_found:
                logger.info(f"ID {input_id} not in value set {e}")
                return False
        return True

    def normalize_identifier(self, input_id: str, cls: ClassDefinition) -> Iterator[str]:
        if self.is_valid_identifier(input_id, cls):
            yield input_id
        for obj_id in self.map_identifier(input_id, cls):
            if obj_id == input_id:
                continue
            if self.is_valid_identifier(obj_id, cls):
                yield obj_id

    def map_identifier(self, input_id: str, cls: ClassDefinition) -> Iterator[str]:
        """
        Normalize an identifier to a preferred prefix.

        :param input_id:
        :param cls:
        :return:
        """
        if input_id.startswith("http://purl.bioontology.org/ontology"):
            # TODO: this should be fixed upstream in OAK
            logging.info(f"Normalizing BioPortal id {input_id}")
            input_id = input_id.replace("http://purl.bioontology.org/ontology/", "").replace(
                "/", ":"
            )
        if input_id.startswith("http://id.nlm.nih.gov/mesh/"):
            # TODO: this should be fixed upstream in OAK
            logging.info(f"Normalizing MESH id {input_id}")
            input_id = input_id.replace("http://id.nlm.nih.gov/mesh/", "").replace("/", ":")
        if input_id.startswith("drugbank:"):
            input_id = input_id.replace("drugbank:", "DRUGBANK:")
        yield input_id
        if not cls.id_prefixes:
            return
        if not self.mappers:
            return
        for mapper in self.mappers:
            if isinstance(mapper, MappingProviderInterface):
                for mapping in mapper.sssom_mappings([input_id]):
                    yield str(mapping.object_id)
            else:
                raise ValueError(f"Unknown mapper type {mapper}")

    def groundings(self, text: str, cls: ClassDefinition) -> Iterator[str]:
        """
        Ground the given text to element identifiers.

        This can potentially yield multiple possible alternatives; these
        should be yielded in priority order.

        - if there is a different singular form of the text, yield from that first
        - dictionary exact matches are yielded first
        - dictionary partial matches are yielded next
        - annotators are yielded next, in order in which they are specified in the schema

        :param text: text to ground, e.g. gene symbol
        :param cls: schema class the ground object should instantiate
        :return:
        """
        logger.info(f"GROUNDING {text} using {cls.name}")
        id_matches = re.match(r"^(\S+):(\d+)$", text)
        if id_matches:
            obj_prefix = id_matches.group(1)
            matching_prefixes = [x for x in cls.id_prefixes if x.upper() == obj_prefix.upper()]
            if matching_prefixes:
                yield matching_prefixes[0] + ":" + id_matches.group(2)
        text_lower = text.lower()
        text_singularized = inflection.singularize(text_lower)
        if text_singularized != text_lower:
            logger.info(f"Singularized {text} to {text_singularized}")
            yield from self.groundings(text_singularized, cls)
        paren_char = "["
        parenthetical_components = re.findall(r"\[(.*?)\]", text_lower)
        if not parenthetical_components:
            paren_char = "("
            parenthetical_components = re.findall(r"\((.*?)\)", text_lower)
        if parenthetical_components:
            logger.info(f"{text_lower} =>paren=> {parenthetical_components}")
            trimmed_text = text_lower
            for component in parenthetical_components:
                if component:
                    logger.debug(
                        f"RECURSIVE GROUNDING OF {component} from {parenthetical_components}"
                    )
                    yield from self.groundings(component, cls)
                if paren_char == "(":
                    trimmed_text = trimmed_text.replace(f"({component})", "")
                elif paren_char == "[":
                    trimmed_text = trimmed_text.replace(f"[{component}]", "")
                else:
                    raise AssertionError(f"Unknown paren char {paren_char}")
            trimmed_text = trimmed_text.strip().replace("  ", " ")
            if trimmed_text:
                if len(trimmed_text) >= len(text_lower):
                    raise AssertionError(
                        f"Trimmed text {trimmed_text} is not shorter than {text_lower}"
                    )
                logger.debug(
                    f"{text_lower} =>trimmed=> {trimmed_text}; in {parenthetical_components}"
                )
                yield from self.groundings(trimmed_text, cls)
        if self.dictionary and text_lower in self.dictionary:
            obj_id = self.dictionary[text_lower]
            logger.debug(f"Found {text} in dictionary: {obj_id}")
            yield obj_id
        if self.dictionary:
            for syn, obj_id in self.dictionary.items():
                if syn in text_lower:
                    if len(syn) / len(text_lower) > self.min_grounding_text_overlap:
                        logger.debug(f"Found {syn} < {text} in dictionary: {obj_id}")
                        yield obj_id
        if self.annotators and cls.name in self.annotators:
            annotators = self.annotators[cls.name]
        else:
            if ANNOTATION_KEY_ANNOTATORS not in cls.annotations:
                annotators = []
            else:
                annotators = cls.annotations[ANNOTATION_KEY_ANNOTATORS].value.split(", ")
        logger.info(f" Annotators: {annotators} [will skip: {self.skip_annotators}]")
        # prioritize whole matches by running these first
        for matches_whole_text in [True, False]:
            config = TextAnnotationConfiguration(matches_whole_text=matches_whole_text)
            for annotator in annotators:
                if isinstance(annotator, str):
                    if self.skip_annotators and annotator in self.skip_annotators:
                        continue
                    if self.annotators is None:
                        self.annotators = {}
                    if annotator not in self.annotators:
                        logger.info(f"Loading annotator {annotator}")
                        self.annotators[annotator] = get_adapter(annotator)
                    annotator = self.annotators[annotator]
                if not matches_whole_text and not isinstance(
                    annotator, OntoPortalImplementationBase
                ):
                    # TODO: allow more fine-grained control
                    logger.info(
                        f"Skipping {type(annotator)} as it does not support partial matches"
                    )
                    continue
                try:
                    results = annotator.annotate_text(text, config)
                    for result in results:
                        yield result.object_id
                except Exception as e:
                    logger.error(f"Error with {annotator} for {text}: {e}")

    # def ground_text_to_id(self, text: str, cls: ClassDefinition = None) -> str:
    #    raise NotImplementedError

    def merge_resultsets(
        self, resultset: List[ExtractionResult], unique_fields: List[str] = None
    ) -> ExtractionResult:
        """
        Merge all resultsets into a single resultset.

        Note the first element of the list is mutated.

        :param resultset:
        :return:
        """
        result = resultset[0].extracted_object
        for next_extraction in resultset[1:]:
            next_result = next_extraction.extracted_object
            if unique_fields:
                for k in unique_fields:
                    if k in result and k in next_result:
                        if result[k] != next_result[k]:
                            logger.error(
                                f"Cannot merge unique fields: {k} {result[k]} != {next_result[k]}"
                            )
                            continue
            for k, v in vars(next_result).items():
                curr_v = getattr(result, k, None)
                if isinstance(v, list):
                    if all(isinstance(x, str) for x in v):
                        setattr(result, k, list(set(curr_v).union(set(v))))
                    else:
                        setattr(result, k, curr_v + v)
                else:
                    if curr_v and v and curr_v != v:
                        logger.error(f"Cannot merge {curr_v} and {v}")
                    if v:
                        setattr(result, k, v)
        return resultset[0]



"""HTML Exporter."""
import html
from dataclasses import dataclass
from pathlib import Path
from typing import Any, TextIO, Union

import pydantic
import yaml

from ontogpt.io.exporter import Exporter, is_curie
from ontogpt.templates.core import ExtractionResult


@dataclass
class HTMLExporter(Exporter):
    """
    An exporter that will generate HTML for extraction results.

    TODO: rewrite to use bootstrap
    """

    output: TextIO = None

    def export(self, extraction_output: ExtractionResult, output: Union[str, Path, TextIO]):
        if isinstance(output, Path):
            output = str(output)
        if isinstance(output, str):
            output = open(str(output), "w", encoding="utf-8")
        self.output = output
        self.export_metadata(extraction_output)
        self.export_results(extraction_output.extracted_object, extraction_output)
        self.h2("YAML Object")
        self.details(yaml.dump(extraction_output.dict()), output, code="yaml")
        self.h2("Prompt")
        self.details(extraction_output.prompt, output)
        self.h2("Completion")
        self.details(extraction_output.raw_completion_output, output)

    def export_metadata(self, extraction_output: ExtractionResult):
        output = self.output
        self.h1(f"Results: {extraction_output.input_id}\n\n")
        self.h2("Input")
        self.i(extraction_output.input_text)

    def export_results(self, obj: Any, extraction_output: ExtractionResult):
        self.h2("Results")
        self.export_object(obj, extraction_output, -1)

    def export_object(
        self, obj: pydantic.BaseModel, extraction_output: ExtractionResult, indent: int
    ):
        self.open_div()
        if indent >= 0:
            self.open_ul()
        for field in obj.__fields__.values():
            if indent < 0:
                self.h3(field.name)
            else:
                self.li(f"<i>{field.name}</i>: ")
            value = getattr(obj, field.name)
            if isinstance(value, pydantic.BaseModel):
                self.export_object(value, extraction_output, indent + 1)
            elif isinstance(value, list):
                self.open_ul()
                n = 1
                for item in value:
                    self.li(f"<b>item: {n}</b>: ")
                    n += 1
                    if isinstance(item, pydantic.BaseModel):
                        self.export_object(
                            item, extraction_output=extraction_output, indent=indent + 1
                        )
                    else:
                        self.export_atom(item, extraction_output, indent + 1)
                self.close_ul()
            else:
                self.export_atom(value, extraction_output, indent + 1)
        if indent >= 0:
            self.open_ul()
        self.close_div()

    def export_atom(self, value, extraction_output: ExtractionResult, indent: int):
        output = self.output
        matches = [
            ne for ne in extraction_output.named_entities if ne.id == value and is_curie(ne.id)
        ]
        if matches:
            match = matches[0]
            output.write(f"{match.label} {self.link(match.id)}")
        else:
            output.write(str(value))
        output.write("\n")

    def details(self, text: str, output: TextIO, code: str = ""):
        output.write("<details>\n")
        output.write("<pre>\n")
        self.w(text)
        output.write("\n</pre>\n")
        output.write("\n</details>\n")

    def link(self, curie: str) -> str:
        return f'<a href="https://bioregistry.io/{curie}">{curie}</a>'

    def open_ul(self):
        self.output.write("<ul>\n")

    def close_ul(self):
        self.output.write("</ul>\n")

    def open_div(self):
        self.output.write("<div>\n")

    def close_div(self):
        self.output.write("</div>\n")

    def li(self, text: str = None):
        self.tag("li", text)

    def h1(self, text: str):
        self.tag("h1", text)

    def h2(self, text: str):
        self.tag("h1", text)

    def h3(self, text: str):
        self.tag("h3", text)

    def i(self, text: str):
        self.tag("i", html.escape(text))

    def tag(self, tag: str, text: str):
        self.output.write(f"<{tag}>{text}</{tag}>\n")

    def w(self, text: str):
        self.output.write(html.escape(text))



"""Evaluate GO."""
from dataclasses import dataclass
from pathlib import Path
from random import shuffle
from typing import Dict, List

import yaml
from oaklib import get_implementation_from_shorthand
from oaklib.datamodels.obograph import LogicalDefinitionAxiom
from oaklib.datamodels.vocabulary import IS_A
from oaklib.interfaces.obograph_interface import OboGraphInterface
from pydantic import BaseModel

from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.evaluation.evaluation_engine import SimilarityScore, SPIRESEvaluationEngine
from ontogpt.templates.metabolic_process import MetabolicProcess

TEST_CASES_DIR = Path(__file__).parent / "test_cases"

METABOLIC_PROCESS = "GO:0008152"
BIOSYNTHESIS = "GO:0009058"
HAS_PRIMARY_OUTPUT = "RO:0004008"


class PredictionGO(BaseModel):
    predicted_object: MetabolicProcess = None
    test_object: MetabolicProcess = None
    scores: Dict[str, SimilarityScore] = None

    def calculate_scores(self):
        self.scores = {}
        for k in ["synonyms", "subclass_of", "inputs", "outputs"]:
            self.scores[k] = SimilarityScore.from_set(
                getattr(self.test_object, k, []),
                getattr(self.predicted_object, k, []),
            )
        for k in ["description"]:
            self.scores[k] = SimilarityScore.from_set(
                getattr(self.test_object, k, "").split(),
                getattr(self.predicted_object, k, "").split(),
            )


class EvaluationObjectSetGO(BaseModel):
    """A result of extracting knowledge on text."""

    test: List[MetabolicProcess] = None
    training: List[MetabolicProcess] = None
    predictions: List[PredictionGO] = None


@dataclass
class EvalGO(SPIRESEvaluationEngine):
    ontology: OboGraphInterface = None
    genus: str = BIOSYNTHESIS
    differentia_relation: str = HAS_PRIMARY_OUTPUT

    def __post_init__(self):
        ontology = get_implementation_from_shorthand("sqlite:obo:go")
        if not isinstance(ontology, OboGraphInterface):
            raise TypeError
        self.ontology = ontology
        self.extractor = SPIRESEngine("metabolic_process.MetabolicProcess")
        self.extractor.labelers = [ontology]

    def make_term_from_ldef(self, ldef: LogicalDefinitionAxiom) -> MetabolicProcess:
        """Make a term from a logical definition."""
        ontology = self.ontology
        term = ldef.definedClassId
        parents = [rel[2] for rel in ontology.relationships([term], [IS_A])]
        mp = MetabolicProcess(
            id=term,
            label=ontology.label(term),
            description=ontology.definition(term),
            synonyms=list(ontology.entity_aliases(term)),
            subclass_of=parents,
        )
        r = ldef.restrictions[0]
        if r.propertyId != HAS_PRIMARY_OUTPUT:
            raise NotImplementedError
        mp.outputs = [r.fillerId]
        return mp

    def valid_test_ids(self) -> List[str]:
        with open(TEST_CASES_DIR / "go-ids-2022.txt") as f:
            return [x.strip() for x in f.readlines()]

    def ldef_matches(self, ldef: LogicalDefinitionAxiom) -> bool:
        """Check if a logical definition matches the genus and differentia."""
        if self.genus not in ldef.genusIds:
            return False
        if len(ldef.restrictions) != 1:
            return False
        if self.differentia_relation != ldef.restrictions[0].propertyId:
            return False
        return True

    def create_test_and_training(
        self, num_test: int = 10, num_training: int = 10
    ) -> EvaluationObjectSetGO:
        """
        Create a test and training set of GO terms.

        This takes around 1m to run.
        """
        ontology = self.ontology
        entities = set(ontology.descendants([self.genus], [IS_A]))
        print(
            f"Found {len(entities)} entities that are descendants of\
                genus {self.genus}; {list(entities)[0:5]}"
        )
        assert "GO:0140872" in entities
        all_test_ids = set(self.valid_test_ids())
        assert "GO:0140872" in all_test_ids
        print(f"Found {len(all_test_ids)} test id candidates; {list(entities)[0:5]}")
        candidate_test_ids = entities.intersection(all_test_ids)
        print(f"Found {len(candidate_test_ids)} candidate test ids")
        assert "GO:0140872" in candidate_test_ids
        candidate_train_ids = entities.difference(all_test_ids)
        print(f"Found {len(candidate_train_ids)} candidate train ids")
        entities = list(candidate_test_ids) + list(candidate_train_ids)
        print(f"Found {len(entities)} entities from {type(ontology)}")
        ldefs = list(ontology.logical_definitions(entities))
        shuffle(ldefs)
        # ldefs = list(ontology.logical_definitions())
        print(f"Found {len(ldefs)} logical definitions")
        ldefs = [ldef for ldef in ldefs if self.ldef_matches(ldef)]
        print(f"Found {len(ldefs)} matching logical definitions")
        ldefs_test = [ldef for ldef in ldefs if ldef.definedClassId in candidate_test_ids]
        print(f"Found {len(ldefs_test)} matching logical definitions for test set")
        ldefs_train = [ldef for ldef in ldefs if ldef.definedClassId not in candidate_test_ids]
        print(f"Found {len(ldefs_train)} matching logical definitions for training set")
        shuffle(ldefs_test)
        shuffle(ldefs_train)
        test = [self.make_term_from_ldef(ldef) for ldef in ldefs_test[:num_test]]
        training = [self.make_term_from_ldef(ldef) for ldef in ldefs_train[:num_training]]

        eos = EvaluationObjectSetGO(test=test, training=training)
        return eos

    def eval(self) -> EvaluationObjectSetGO:
        ke = self.extractor
        eos = self.create_test_and_training()
        eos.predictions = []
        print(yaml.dump(eos.dict()))
        for test_obj in eos.test[0:10]:
            print(yaml.dump(test_obj.dict()))
            predicted_obj = ke.generalize({"label": test_obj.label}, eos.training[0:4])
            pred = PredictionGO(predicted_object=predicted_obj, test_object=test_obj)
            pred.calculate_scores()
            eos.predictions.append(pred)
        return eos



"""oak-ai package."""
from importlib import metadata

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover



"""Webapp main function."""
from io import StringIO
from pathlib import Path
from typing import Dict

import uvicorn
from fastapi import FastAPI, Form, Request
from pydantic import BaseModel
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from ontogpt.engines.enrichment import EnrichmentEngine
from ontogpt.engines.knowledge_engine import DATAMODELS, MODELS
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.io.html_exporter import HTMLExporter
from ontogpt.utils.gene_set_utils import GeneSet

this_path = Path(__file__).parent
static_dir = this_path / "static"
html_dir = this_path / "html"


class Query(BaseModel):
    text: str
    datamodel: str


app = FastAPI()
static_dir.mkdir(parents=True, exist_ok=True)
app.mount("/static", StaticFiles(directory=str(static_dir)), name="static")
templates = Jinja2Templates(directory=str(html_dir))

html_exporter = HTMLExporter()
engines: Dict[str, SPIRESEngine] = {}


def get_engine(datamodel: str):
    if datamodel not in engines:
        engines[datamodel] = SPIRESEngine(datamodel)
    return engines[datamodel]


@app.get("/")
def read_root(request: Request):
    return templates.TemplateResponse(
        "form.html", context={"request": request, "datamodels": DATAMODELS}
    )


@app.post("/")
def form_post(request: Request, datamodel: str = Form(...), text: str = Form(...)):
    print(f"Received request with schema {datamodel}")
    print(f"Received request with text {text}")
    engine = get_engine(datamodel)
    ann = engine.extract_from_text(text)
    print(f"Got {ann}")
    output = StringIO()
    html_exporter.export(ann, output)
    return templates.TemplateResponse(
        "results.html", context={"request": request, "inner_html": output.getvalue()}
    )


@app.get("/spindoctor")
def sd_read_root(request: Request):
    return templates.TemplateResponse(
        "spindoctor/form.html", context={"request": request, "models": MODELS}
    )


@app.post("/spindoctor")
def sd_form_post(request: Request, model: str = Form(...), text: str = Form(...)):
    print(f"Received request with model {model}")
    print(f"Received request with text {text}")
    symbols = [s.strip() for s in text.split("\n")]
    engine = EnrichmentEngine(model=model)
    gene_set = GeneSet(name="TEMP", gene_symbols=symbols)
    ann = engine.summarize(gene_set)
    print(f"Got {ann}")
    output = StringIO()
    html_exporter.export(ann, output)
    return templates.TemplateResponse(
        "spindoctor/results.html", context={"request": request, "inner_html": output.getvalue()}
    )


def start():
    """Launch with `poetry run start` at root level."""
    uvicorn.run("ontogpt.webapp.main:app", host="127.0.0.1", port=8000, reload=True)



from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from linkml_runtime.linkml_model import Decimal
from pydantic import BaseModel as BaseModel
from pydantic import Field

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = "__weakref__"


class ConfiguredBaseModel(
    WeakRefShimBaseModel,
    validate_assignment=True,
    validate_all=True,
    underscore_attrs_are_private=True,
    extra="forbid",
    arbitrary_types_allowed=True,
):
    pass


class ProblemType(str, Enum):
    UNKNOWN = "UNKNOWN"
    TYPO = "TYPO"
    BAD_XREF = "BAD_XREF"
    MISSING_DEFINITION = "MISSING_DEFINITION"
    MISSING_PARENT = "MISSING_PARENT"
    BAD_RELATIONSHIP = "BAD_RELATIONSHIP"
    OTHER = "OTHER"


class ChangeType(str, Enum):
    UNKNOWN = "UNKNOWN"
    CHANGE_DEFINITION = "CHANGE_DEFINITION"
    CHANGE_LABEL = "CHANGE_LABEL"
    CHANGE_XREF = "CHANGE_XREF"
    MOVE_TERM = "MOVE_TERM"


class OntologyIssue(ConfiguredBaseModel):
    title: Optional[str] = Field(None, description="""the title of the issue""")
    summary: Optional[str] = Field(None, description="""a high level summary""")
    status: Optional[str] = Field(None)
    domains: Optional[List[str]] = Field(
        default_factory=list, description="""What part of the ontology does this pertain to."""
    )
    problem_list: Optional[List[OntologyProblem]] = Field(
        default_factory=list, description="""A list of problems stated at a high level"""
    )
    proposed_changes: Optional[List[OntologyChange]] = Field(
        default_factory=list, description="""What part of the ontology does this pertain to."""
    )


class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """

    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(
        None, description="""The complex objects extracted from the text"""
    )
    named_entities: Optional[List[Any]] = Field(
        default_factory=list, description="""Named entities extracted from the text"""
    )


class NamedEntity(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class OntologyClass(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class CompoundExpression(ConfiguredBaseModel):
    None


class OntologyProblem(CompoundExpression):
    description: Optional[str] = Field(
        None, description="""A succinct description of the problem"""
    )
    severity: Optional[str] = Field(None, description="""How severe is this problem?""")
    category: Optional[ProblemType] = Field(
        None, description="""What category does this problem fall into?"""
    )
    about: Optional[List[str]] = Field(
        default_factory=list, description="""What terms in the ontology is this problem about?"""
    )


class OntologyChange(CompoundExpression):
    description: Optional[str] = Field(
        None, description="""A succinct description of the proposed change"""
    )
    category: Optional[ChangeType] = Field(None, description="""What kind of change?""")
    about: Optional[List[str]] = Field(
        default_factory=list, description="""What terms in the ontology will this change affect?"""
    )


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """

    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(
        None, description="""A qualifier for the statements, e.g. \"NOT\" for negation"""
    )
    subject_qualifier: Optional[str] = Field(
        None,
        description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""",
    )
    object_qualifier: Optional[str] = Field(
        None,
        description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""",
    )


class TextWithTriples(ConfiguredBaseModel):
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)


class RelationshipType(NamedEntity):
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Publication(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")


class AnnotatorResult(ConfiguredBaseModel):
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
OntologyIssue.update_forward_refs()
ExtractionResult.update_forward_refs()
NamedEntity.update_forward_refs()
OntologyClass.update_forward_refs()
CompoundExpression.update_forward_refs()
OntologyProblem.update_forward_refs()
OntologyChange.update_forward_refs()
Triple.update_forward_refs()
TextWithTriples.update_forward_refs()
RelationshipType.update_forward_refs()
Publication.update_forward_refs()
AnnotatorResult.update_forward_refs()



"""DrugMechDB."""
from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel as BaseModel
from pydantic import Field

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = "__weakref__"


class ConfiguredBaseModel(
    WeakRefShimBaseModel,
    validate_assignment=True,
    validate_all=True,
    underscore_attrs_are_private=True,
    extra="forbid",
    arbitrary_types_allowed=True,
):
    pass


class Mechanism(ConfiguredBaseModel):
    directed: Optional[bool] = Field(None)
    multigraph: Optional[bool] = Field(None)
    reference: Optional[List[str]] = Field(default_factory=list)
    references: Optional[str] = Field(None)
    comment: Optional[str] = Field(None)
    comments: Optional[str] = Field(None)
    commments: Optional[str] = Field(None)
    graph: Optional[Graph] = Field(None)
    links: Optional[List[Link]] = Field(default_factory=list)
    nodes: Optional[List[Node]] = Field(default_factory=list)


class Graph(ConfiguredBaseModel):
    id: Optional[str] = Field(None)
    disease: Optional[str] = Field(None)
    disease_mesh: Optional[str] = Field(None)
    drug: Optional[str] = Field(None)
    drug_mesh: Optional[str] = Field(None)
    drugbank: Optional[str] = Field(None)
    alt_ids: Optional[str] = Field(None)


class Link(ConfiguredBaseModel):
    key: Optional[str] = Field(None)
    source: Optional[str] = Field(None)
    target: Optional[str] = Field(None)
    reference: Optional[str] = Field(None)


class Node(ConfiguredBaseModel):
    id: Optional[str] = Field(None)
    label: Optional[str] = Field(None)
    name: Optional[str] = Field(None)
    all_id: Optional[str] = Field(None)
    alt_ids: Optional[List[str]] = Field(default_factory=list)
    alt_names: Optional[List[str]] = Field(default_factory=list)
    alt_name: Optional[str] = Field(None)
    reference: Optional[str] = Field(None)


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
Mechanism.update_forward_refs()
Graph.update_forward_refs()
Link.update_forward_refs()
Node.update_forward_refs()



"""
DrugMechDB evaluation.

Source Data Model: a direct transliteration of the DrugMechDB schema
Target Data Model: the semllama native representation of drugs and mechanism
"""
import gzip
import logging
from dataclasses import dataclass
from pathlib import Path
from random import shuffle
from typing import Any, Dict, List, Optional, TextIO, Union

import yaml
from oaklib import get_implementation_from_shorthand
from pydantic import BaseModel

import ontogpt.evaluation.drugmechdb.datamodel.drugmechdb as source_datamodel
import ontogpt.templates.drug as target_datamodel
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.evaluation.evaluation_engine import SimilarityScore, SPIRESEvaluationEngine

THIS_DIR = Path(__file__).parent
DATABASE_DIR = Path(__file__).parent / "database"
MOA_TEXTS = DATABASE_DIR / "drugbank-to-mechanism-text.tsv"
TEST_CASES_DIR = THIS_DIR / "test_cases"
EXEMPLARS_DIR = THIS_DIR / "exemplars"
EXEMPLAR_CASES = EXEMPLARS_DIR / "drugmechdb-exemplars.yaml"


def _fix_source_mechanism(mechanism_dict: dict) -> dict:
    """Fix the id field in Graph objects.

    drugmechdb uses `_id` in the json but this is not
    compatible with pydantic, so we translate to `id`

    https://github.com/linkml/linkml/issues/1179
    """
    g = mechanism_dict["graph"]
    g["id"] = g["_id"]
    del g["_id"]
    # normalize alt_ids
    bad_fields = ["all_id", "alt_name", "alt-name", "comemt", "comemnt"]
    for n in mechanism_dict["nodes"]:
        if "alt_ids" in n and isinstance(n["alt_ids"], str):
            n["alt_ids"] = [n["alt_ids"]]
        for f in bad_fields:
            if f in n:
                del n[f]
                # don't bother trying to repair for now - too messy
    for f in bad_fields:
        if f in mechanism_dict:
            del mechanism_dict[f]
    # print(mechanism_dict)
    return mechanism_dict


def _fix_id(id: Optional[str]) -> str:
    return id.replace("UniProt:", "PR:").replace("DB:", "drugbank:") if id else None


class PredictionDrugMechDB(BaseModel):
    predicted_object: target_datamodel.DrugMechanism = None
    test_object: target_datamodel.DrugMechanism = None
    scores: Dict[str, SimilarityScore] = None
    named_entities: List[Any] = None

    def calculate_scores(self):
        self.scores = {}

        def all_objects(dm: target_datamodel.DrugMechanism):
            return list(
                set(link.subject for link in dm.mechanism_links)
                | set(link.object for link in dm.mechanism_links)
            )

        self.scores["similarity"] = SimilarityScore.from_set(
            all_objects(self.test_object),
            all_objects(self.predicted_object),
        )


class EvaluationObjectSetDrugMechDB(BaseModel):
    """A result of predicting paths."""

    test: List[target_datamodel.DrugMechanism] = None
    training: List[target_datamodel.DrugMechanism] = None
    predictions: List[PredictionDrugMechDB] = None


@dataclass
class EvalDrugMechDB(SPIRESEvaluationEngine):
    # ontology: OboGraphInterface = None
    data: List[target_datamodel.DrugMechanism] = None
    _drug_to_mechanism_text: Dict[str, str] = None
    default_labelers = [
        "sqlite:obo:mesh",
        "sqlite:obo:drugbank",
        "sqlite:obo:go",
        "sqlite:obo:uberon",
        "sqlite:obo:pr",
    ]

    def __post_init__(self):
        self.extractor = SPIRESEngine("drug.DrugMechanism")
        self.extractor.labelers = [
            get_implementation_from_shorthand(lab) for lab in self.default_labelers
        ]

    @property
    def drug_to_mechanism_text(self) -> Dict[str, str]:
        """Mapping between drugbank id and MOA text from drugbank record.

        :return:
        """
        if self._drug_to_mechanism_text is None:
            with open(MOA_TEXTS, "r") as f:
                self._drug_to_mechanism_text = {}
                for line in f:
                    toks = line.strip().split("\t")
                    if len(toks) != 2:
                        logging.warning(f"Bad line: {line}")
                        continue
                    drug, text = toks
                    if text != "NA":
                        self._drug_to_mechanism_text[f"drugbank:{drug}"] = text
        return self._drug_to_mechanism_text

    def load_source_mechanisms_from_path(
        self, file: Union[str, Path, TextIO]
    ) -> List[source_datamodel.Mechanism]:
        if isinstance(file, Path):
            file = str(file)
        if isinstance(file, str):
            with open(file, "r") as f:
                return self.load_source_mechanisms_from_path(f)
        mechanisms = yaml.safe_load(file)
        print(f"Loading {len(mechanisms)} mechanism objects from yaml; translating...")
        return [source_datamodel.Mechanism(**_fix_source_mechanism(m)) for m in mechanisms]

    def load_exemplars(self) -> List[target_datamodel.DrugMechanism]:
        srcs = self.load_source_mechanisms_from_path(EXEMPLAR_CASES)
        return [self.transform_mechanism(m) for m in srcs]

    def load_and_transform_source_database(self) -> List[target_datamodel.DrugMechanism]:
        """Load the entire DrugMechDB database."""
        with gzip.open(str(DATABASE_DIR / "indication_paths.yaml.gz"), "rb") as f:
            srcs = self.load_source_mechanisms_from_path(f)
            return [self.transform_mechanism(m) for m in srcs]

    def load_target_database(self) -> List[target_datamodel.DrugMechanism]:
        """Load the entire DrugMechDB database."""
        with gzip.open(str(DATABASE_DIR / "drugmechdb-normalized.yaml.gz"), "rb") as f:
            objs = yaml.safe_load(f)
            return [target_datamodel.DrugMechanism(**m) for m in objs]

    def transform_mechanism(
        self, mechanism: source_datamodel.Mechanism
    ) -> target_datamodel.DrugMechanism:
        """Translate a mechanism from DrugMechDB to the target template."""
        triples = []
        for link in mechanism.links:
            triples.append(
                target_datamodel.MechanismLink(
                    subject=_fix_id(link.source),
                    object=_fix_id(link.target),
                    predicate=link.key,
                )
            )
        # the source YAML is a little odd;
        # - the singular field "reference" is usually a *list*
        # - the field "references" is usually a single string
        refs = mechanism.reference if mechanism.reference else []
        if mechanism.references:
            refs.append(mechanism.references)
        return target_datamodel.DrugMechanism(
            disease=mechanism.graph.disease_mesh,
            drug=_fix_id(mechanism.graph.drugbank),
            mechanism_links=triples,
            references=refs,
        )

    def create_test_and_training(
        self, num_test: int = 10, num_training: int = 5, include_texts: bool = False
    ) -> EvaluationObjectSetDrugMechDB:
        """Create a test and training set from the DrugMechDB database."""
        if self.data is None:
            mechanisms = self.load_target_database()
        else:
            mechanisms = self.data
        if include_texts:
            mechanisms = [m for m in mechanisms if m.drug in self.drug_to_mechanism_text]
        shuffle(mechanisms)
        return EvaluationObjectSetDrugMechDB(
            test=mechanisms[:num_test],
            training=mechanisms[num_test : num_test + num_training],
        )

    def create_training_set(self, num=100):
        ke = self.extractor
        mechanisms = self.load_target_database()
        mechanisms = [m for m in mechanisms if m.drug in self.drug_to_mechanism_text]
        shuffle(mechanisms)
        for m in mechanisms[0:num]:
            stub = {
                "disease": m.disease,
                "drug": m.drug,
            }
            text = self.drug_to_mechanism_text[m.drug]
            prompt = ke.get_completion_prompt(None, text, object=stub)
            completion = ke.serialize_object(m)
            yield dict(prompt=prompt, completion=completion)

    def eval(self) -> EvaluationObjectSetDrugMechDB:
        """Evaluate the ability to extract a path from MOA text."""
        num_test = self.num_tests
        ke = self.extractor
        mechanisms = self.load_target_database()

        def is_candidate(m: target_datamodel.DrugMechanism):
            if m.drug not in self.drug_to_mechanism_text:
                return False
            if len(m.references) != 1:
                return False
            ref = m.references[0]
            if ref.startswith("https://go.drugbank.com/drugs/") and ref.endswith(
                "#mechanism-of-action"
            ):
                return True
            return False

        mechanisms = [m for m in mechanisms if is_candidate(m)]
        print(f"Using {len(mechanisms)} mechanisms")
        shuffle(mechanisms)
        eos = EvaluationObjectSetDrugMechDB(
            test=mechanisms[:num_test],
            training=[],
            predictions=[],
        )
        print(yaml.dump(eos.dict()))
        for test_obj in eos.test:
            text = self.drug_to_mechanism_text[test_obj.drug]
            test_obj.source_text = text
            print(yaml.dump(test_obj.dict()))
            stub = {
                "disease": test_obj.disease,
                "drug": test_obj.drug,
            }
            results = ke.extract_from_text(text, object=stub)
            predicted_obj = results.extracted_object
            pred = PredictionDrugMechDB(predicted_object=predicted_obj, test_object=test_obj)
            pred.named_entities = results.named_entities
            print(yaml.dump(pred.dict()))
            pred.calculate_scores()
            print(yaml.dump(pred.dict()))
            eos.predictions.append(pred)
        return eos

    def eval_path_prediction(self) -> EvaluationObjectSetDrugMechDB:
        """Evaluate the ability to predict a path purely from background knowledge in the LLM.

        :return:
        """
        ke = self.extractor
        eos = self.create_test_and_training(num_test=self.num_tests, num_training=self.num_training)
        eos.predictions = []
        print(yaml.dump(eos.dict()))
        for test_obj in eos.test:
            print(yaml.dump(test_obj.dict()))
            stub = {
                "disease": test_obj.disease,
                "drug": test_obj.drug,
            }
            results = ke.generalize(stub, eos.training)
            predicted_obj = results.extracted_object[0]
            pred = PredictionDrugMechDB(predicted_object=predicted_obj, test_object=test_obj)
            pred.calculate_scores()
            eos.predictions.append(pred)
        return eos



"""Utilities for handling gene sets."""
import csv
import glob
import json
import logging
from copy import deepcopy
from pathlib import Path
from random import sample
from typing import Dict, List, Optional, Tuple, Union

import requests_cache
import yaml
from oaklib import BasicOntologyInterface, get_adapter
from pydantic import BaseModel

ENTITY_ID = str
SYMBOL = str
DESCRIPTION = str
GENE_TUPLE = Tuple[ENTITY_ID, SYMBOL, DESCRIPTION]

GENE_REQUESTS_CACHE = ".gene_requests_cache"


logger = logging.getLogger(__name__)

session = requests_cache.CachedSession(GENE_REQUESTS_CACHE)


class Gene(BaseModel):
    id: str
    symbol: str = None
    ontological_synopsis: str = None
    narrative_synopsis: str = None


class GeneSet(BaseModel):
    """A set of genes."""

    name: str
    gene_symbols: Optional[List[str]] = None
    gene_ids: Optional[List[str]] = None
    genes: Optional[Dict[str, Gene]] = None
    taxon: str = "human"
    taxon_id: Optional[str] = None
    description: Optional[str] = None
    source: Optional[str] = None
    source_url: Optional[str] = None
    target_term_ids: Optional[List[str]] = None
    number_of_genes_dropped: Optional[int] = None
    original_name: Optional[str] = None
    lacks_descriptions: Optional[bool] = None


class GeneSetCollection(BaseModel):
    """A collection of gene sets."""

    gene_sets: List[GeneSet]


def save_gene_set(gene_set: GeneSet, path: Union[str, Path]):
    """
    Save a gene set to a file.

    :param gene_set:
    :param path:
    :return:
    """
    if isinstance(path, Path):
        path = str(path)
    with open(path, "w", encoding="utf-8") as f:
        f.write(yaml.dump(gene_set.dict(exclude_unset=True), sort_keys=False))


def parse_gene_set(input_path: Union[str, Path], format: str = None) -> GeneSet:
    """
    Parse a gene set from a file.

    Accepts:

    - yaml (native gene set format; recommended)
    - json (msigdb format)
    - txt (one gene symbol per line)

    This does no normalization of gene symbols or ids.
    """
    if isinstance(input_path, Path):
        input_path = str(input_path)
    if format is None:
        if "gene_export_geneset" in input_path:
            format = "geneweaver"
        else:
            format = input_path.split(".")[-1]
    if format == "yaml":
        with open(input_path, "r") as f:
            gene_set = GeneSet(**yaml.safe_load(f))
    elif format == "json" or format == "msigdb":
        with open(input_path, "r") as f:
            name, msig = list(json.load(f).items())[0]
            gene_set = GeneSet(name=name, gene_symbols=msig["geneSymbols"])
    elif format == "geneweaver":
        gene_set = parse_geneweaver(input_path)
    elif format == "txt":
        with open(input_path, "r") as f:
            gene_symbols = [line.strip() for line in f.readlines()]
            gene_set = GeneSet(name=input_path, gene_symbols=gene_symbols)
    else:
        raise ValueError(f"Unknown format {format}")
    return gene_set


def parse_geneweaver(path: str) -> GeneSet:
    gene_set = GeneSet(name=Path(path).stem, source="geneweaver", gene_ids=[], gene_symbols=[])
    col2sp = {
        "Wormbase": "C elegans",
        "ZFIN": "Danio rerio",
    }
    fix = {
        "Wormbase": "WB",
    }
    with open(path, "r") as f:
        reader = csv.DictReader(f, delimiter="\t")
        col = None
        sp = None
        for row in reader:
            sym = row["Gene Symbol"]
            if not col:
                for k, v in col2sp.items():
                    if row.get(k, None):
                        col = k
                        db = col
                        gene_set.taxon = v
                        if col in fix:
                            db = fix[col]
                        break
            if not col:
                raise ValueError("Unknown column")

            id = f"{db}:{row[col]}"
            gene = Gene(id=id, symbol=sym)
            gene_set.gene_ids.append(id)
            gene_set.gene_symbols.append(sym)
            gene_set.gene_ids.append(id)
    return gene_set



def load_gene_sets(
    path: str, ontology_adapter: BasicOntologyInterface = None, strict=False, fill_missing=True,
) -> GeneSetCollection:
    """Load collection of gene sets from a folder.

    If ontology_adapter is provided, gene symbols will be converted to gene ids and vice versa.
    """
    gene_sets = []
    for input_path in glob.glob(f"{path}/*.yaml"):
        gene_set = parse_gene_set(input_path)
        gene_sets.append(gene_set)
        if not gene_set.gene_ids and not gene_set.gene_symbols:
            raise ValueError(f"Gene set {gene_set.name} has no gene symbols or ids")
        if fill_missing:
            fill_missing_gene_set_values(gene_set, ontology_adapter, strict)
    return GeneSetCollection(gene_sets=gene_sets)


def _is_human(gene_set: GeneSet) -> bool:
    """
    Check if a gene set is human.

    Note we need to handle this as a special case due to the need to
    depend on sqlite:obo:hgnc

    :param gene_set:
    :return:
    """
    if gene_set.gene_ids:
        if all([id.startswith("HGNC:") for id in gene_set.gene_ids]):
            return True
        if any([id.startswith("HGNC:") for id in gene_set.gene_ids]):
            raise ValueError("Gene set contains mixed taxons. Please use a different gene set.")
        return False
    return gene_set.taxon.lower() == "human" or gene_set.taxon_id == "NCBITaxon:9606"


def fill_missing_gene_set_values(
    gene_set: GeneSet,
    ontology_adapter: BasicOntologyInterface = None,
    strict=False,
):
    if not gene_set.gene_ids:
        # special purpose code to fill in HGNC symbols
        if not ontology_adapter:
            if not _is_human(gene_set):
                raise ValueError("Gene set is not human and no ontology adapter was provided")
            ontology_adapter = get_adapter("sqlite:obo:hgnc")
            logger.info(f"Fetching ids for {len(gene_set.gene_symbols)} genes")
            gene_set.gene_ids = []
            for sym in gene_set.gene_symbols:
                ids = ontology_adapter.curies_by_label(sym)
                if strict and len(ids) != 1:
                    raise ValueError(f"Could not find a single id for symbol {sym}")
                # hack! normalize casing
                gene_set.gene_ids.extend([id.upper() for id in ids])
    for gene_id in gene_set.gene_ids:
        # fill each gene using alliance API
        if not gene_set.genes:
            gene_set.genes = {}
        if gene_id not in gene_set.genes:
            gene_set.genes[gene_id] = Gene(id=gene_id)
        gene = gene_set.genes[gene_id]
        if gene_set.lacks_descriptions and gene.symbol:
            continue
        if not gene.symbol or gene.ontological_synopsis is None or gene.narrative_synopsis is None:
            symbol, narrative_synopsis, ontological_synopsis = gene_info(gene_id)
            if not gene.symbol:
                gene.symbol = symbol
            if not gene.ontological_synopsis:
                gene.ontological_synopsis = ontological_synopsis
            if not gene.narrative_synopsis:
                gene.narrative_synopsis = narrative_synopsis
    if not gene_set.gene_symbols:
        gene_set.gene_symbols = [gene.symbol for gene in gene_set.genes.values()]


def drop_genes_from_gene_set(gene_set: GeneSet, percent_to_drop: int) -> GeneSet:
    """
    Drop genes from a gene set.

    :param gene_set:
    :param percent_to_drop:
    :return:
    """
    gene_set = deepcopy(gene_set)
    if percent_to_drop < 0 or percent_to_drop > 100:
        raise ValueError("Percent to drop must be between 0 and 100")
    if percent_to_drop == 0:
        return gene_set
    num_genes = len(gene_set.genes)
    if not num_genes:
        raise ValueError("Gene set has no genes")
    num_to_drop = int(num_genes * percent_to_drop / 100)
    if num_to_drop == 0:
        num_to_drop = 1
    logger.info(f"Dropping {num_to_drop} genes from {gene_set.name}")
    gene_set.genes = dict(sample(gene_set.genes.items(), num_genes - num_to_drop))
    gene_set.gene_ids = list(gene_set.genes.keys())
    gene_set.gene_symbols = [gene.symbol for gene in gene_set.genes.values()]
    gene_set.number_of_genes_dropped = num_to_drop
    gene_set.original_name = gene_set.name
    return gene_set


def gene_info(id: ENTITY_ID) -> Tuple[SYMBOL, DESCRIPTION, DESCRIPTION]:
    url = f"https://www.alliancegenome.org/api/gene/{id}"
    logger.info(f"Fetching gene summary from {url}")
    response = session.get(url)
    if response.status_code != 200:
        raise ValueError(f"Error fetching issues: {response.status_code} // {response.text}")
    obj = response.json()
    symbol = obj["symbol"]
    return symbol, obj["geneSynopsis"], obj["automatedGeneSynopsis"]



from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union, Literal
from pydantic import BaseModel as BaseModel, Field
from linkml_runtime.linkml_model import Decimal

metamodel_version = "None"
version = "None"

class WeakRefShimBaseModel(BaseModel):
   __slots__ = '__weakref__'
    
class ConfiguredBaseModel(WeakRefShimBaseModel,
                validate_assignment = True, 
                validate_all = True, 
                underscore_attrs_are_private = True, 
                extra = 'forbid', 
                arbitrary_types_allowed = True):
    pass                    


class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    


class NamedEntity(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class DiagnosticProcedure(NamedEntity):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class Phenotype(NamedEntity):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class ClinicalAttribute(NamedEntity):
    
    unit: Optional[str] = Field(None, description="""the unit used to measure the attribute""")
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class Quality(NamedEntity):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class Unit(NamedEntity):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class CompoundExpression(ConfiguredBaseModel):
    
    None
    


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    


class DiagnosticProceduretoPhenotypeAssociation(Triple):
    """
    A triple representing a relationship between a diagnostic procedure and an associated phenotype, e.g., \"blood pressure measurement\" is associated with \"high blood pressure\".
    """
    subject: Optional[str] = Field(None, description="""A diagnostic procedure yielding a result, which in turn may be interpreted as a phenotype. Procedures include \"heart rate measurement\", \"blood pressure measurement\", \"oxygen saturation measurement\", etc. In practice, procedures may be named based on what they measure, with the \"measurement\" part left implicit.""")
    predicate: Optional[str] = Field(None, description="""The relationship type, e.g. RELATED_TO""")
    object: Optional[List[str]] = Field(default_factory=list, description="""The observable physical or biochemical characteristics of a patient. Not equivalent to a disease state, but may contribute to a diagnosis.""")
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the procedure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the phenotype.""")
    


class DiagnosticProceduretoAttributeAssociation(Triple):
    """
    A triple representing a relationship between a diagnostic procedure and a measured attribute, e.g., \"blood pressure measurement\" is associated with \"blood pressure\" (or in OBA, something like OBA:VT0000183, \"blood pressure trait\").
    """
    subject: Optional[str] = Field(None, description="""A diagnostic procedure yielding a result, which in turn may be interpreted as a phenotype. Procedures include \"heart rate measurement\", \"blood pressure measurement\", \"oxygen saturation measurement\", etc. In practice, procedures may be named based on what they measure, with the \"measurement\" part left implicit.""")
    predicate: Optional[str] = Field(None, description="""The relationship type, e.g. RELATED_TO""")
    object: Optional[List[str]] = Field(default_factory=list, description="""Any measurable clinical attribute.""")
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the procedure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the phenotype.""")
    


class TextWithTriples(ConfiguredBaseModel):
    
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    


class RelationshipType(NamedEntity):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class ProcedureToPhenotypePredicate(RelationshipType):
    """
    A predicate for procedure to phenotype relationships, defining \"this procedure is intended to provide support for/against this phenotype\".
    """
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class ProcedureToAttributePredicate(RelationshipType):
    """
    A predicate for procedure to attribute relationships, defining \"this procedure is a measurement of this attribute\".
    """
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    


class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    


class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    



# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
ExtractionResult.update_forward_refs()
NamedEntity.update_forward_refs()
DiagnosticProcedure.update_forward_refs()
Phenotype.update_forward_refs()
ClinicalAttribute.update_forward_refs()
Quality.update_forward_refs()
Unit.update_forward_refs()
CompoundExpression.update_forward_refs()
Triple.update_forward_refs()
DiagnosticProceduretoPhenotypeAssociation.update_forward_refs()
DiagnosticProceduretoAttributeAssociation.update_forward_refs()
TextWithTriples.update_forward_refs()
RelationshipType.update_forward_refs()
ProcedureToPhenotypePredicate.update_forward_refs()
ProcedureToAttributePredicate.update_forward_refs()
Publication.update_forward_refs()
AnnotatorResult.update_forward_refs()




"""OWL convertor."""
from dataclasses import dataclass
from pathlib import Path
from typing import TextIO, Union

import pydantic
from linkml.generators.pythongen import PythonGenerator
from linkml_owl.dumpers.owl_dumper import OWLDumper
from linkml_runtime import SchemaView

from ontogpt.io.exporter import Exporter
from ontogpt.templates.core import ExtractionResult


@dataclass
class OWLExporter(Exporter):
    def export(
        self,
        extraction_output: ExtractionResult,
        output: Union[str, Path, TextIO],
        schemaview: SchemaView,
        id_value=None,
    ):
        if isinstance(output, Path):
            output = str(output)
        if isinstance(output, str):
            output = open(str(output), "w", encoding="utf-8")
        dumper = OWLDumper()
        element = extraction_output.extracted_object
        cls_name = type(element).__name__
        id_slot = schemaview.get_identifier_slot(cls_name)
        if id_slot is not None:
            id_slot_name = id_slot.name
            if id_value is not None:
                setattr(element, id_slot_name, "AUTO:_ROOT")
            else:
                id_value = getattr(element, id_slot_name, None)
                print(f"Setting {id_slot_name} [{id_value}] to AUTO:_ROOT for {cls_name}")
                if id_value is None:
                    setattr(element, id_slot_name, "AUTO:_ROOT")
        axioms = []
        for named_entity in extraction_output.named_entities:
            ne_as_dc = self._as_dataclass_object(named_entity, schemaview)
            doc = dumper.to_ontology_document(ne_as_dc, schemaview.schema)
            axioms.extend(doc.ontology.axioms)
        element_as_dataclass = self._as_dataclass_object(element, schemaview)
        doc = dumper.to_ontology_document(element_as_dataclass, schemaview.schema)
        doc.ontology.axioms.extend(axioms)
        output.write(str(doc))

    def _as_dataclass_object(self, element: pydantic.BaseModel, schemaview: SchemaView):
        cls_name = type(element).__name__
        schemaview.merge_imports()
        element_dict = element.dict()
        dataclasses_module = PythonGenerator(schemaview.schema).compile_module()
        target_class_py = dataclasses_module.__dict__[cls_name]
        element_as_dataclass = target_class_py(**element_dict)
        return element_as_dataclass



"""Enrichment engine."""
import logging
import re
from collections.abc import Iterator
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from jinja2 import Template
from oaklib import BasicOntologyInterface, get_adapter
from pydantic import BaseModel

from ontogpt.engines.knowledge_engine import KnowledgeEngine, MODEL_GPT_4
from ontogpt.prompts.enrichment import DEFAULT_ENRICHMENT_PROMPT
from ontogpt.templates.class_enrichment import ClassEnrichmentResult
from ontogpt.templates.gene_description_term import GeneDescriptionTerm
from ontogpt.utils.gene_set_utils import (
    ENTITY_ID,
    GENE_TUPLE,
    GeneSet,
    fill_missing_gene_set_values,
)

logger = logging.getLogger(__name__)


SUMMARY_KEYWORD = "Summary"
MECHANISM_KEYWORD = "Mechanism"
ENRICHED_TERMS_KEYWORD = "Enriched Terms"


class GeneDescriptionSource(Enum):
    NONE = "none"
    ONTOLOGICAL_SYNOPSIS = "ontological"
    NARRATIVE_SYNOPSIS = "narrative"
    COMBINED_SYNOPSIS = "combined"
    WIKIPEDIA = "wikipedia"
    PUBMED = "pubmed"


class EnrichmentPayload(BaseModel):
    """Payload for enrichment."""

    prompt: str = None
    """The prompt to use for the summarization task (only filled for LLMs)."""

    response_text: str = None
    """The response text from the summarization task (only filled for LLMs)."""

    truncation_factor: float = None
    """Fraction of gene descriptions retained after trimming to fit token limit."""

    summary: str = None
    """The summary of the gene set generated by the LLM."""

    term_strings: List[str] = []
    """The raw term labels parsed from the LLM completion payload"""

    term_ids: List[str] = []
    """The normalized terms"""

    ontological_synopsis: bool = None
    """True if the gene descriptions used the ontological synopsis"""

    combined_synopsis: bool = None
    """True if the gene descriptions used both ontological and narrative synopses"""

    annotations: bool = None
    """True if the gene descriptions used the annotations (vs latent KB)"""

    response_token_length: int = None
    """The number of tokens in the response text"""

    model: str = None
    """The model used for the summarization task (only filled for LLMs)."""

    method: str = None
    """The method used for the summarization task."""

    prompt_variant: str = None

    enrichment_results: List[ClassEnrichmentResult] = None
    """Enrichment results (only filled for non-LLMs)"""


@dataclass
class EnrichmentEngine(KnowledgeEngine):
    """Engine for performing term enrichment.

    Algorithm: SPINDOCTOR

    Summarizing Proteins Interpolating Narrative
    Descriptions of Controlled Terms for Ontological enRichment

    """

    # engine: str = "text-davinci-003"
    # model: str = "xxxgpt-3.5-turbo"

    label_resolvers: Dict[str, BasicOntologyInterface] = field(default_factory=dict)

    completion_length = 250

    prompt_template: str = None
    """Path to a jinja2 template for the prompt"""

    end_marker = "###"

    def __post_init__(self):
        if not self.template:
            self.template = "gene_description_term"
        super().__post_init__()

    def __key(self):
        return "enrichment_engine"

    def __eq__(self, other):
        if isinstance(other, EnrichmentEngine):
            return self.__key() == other.__key()
        return NotImplemented

    def __hash__(self):
        return hash(self.__key())

    def summarize(
        self,
        gene_set: GeneSet,
        prompt_template: Optional[str] = None,
        normalize=False,
        strict=False,
        gene_description_source: GeneDescriptionSource = None,
        ontological_synopsis=True,
        combined_synopsis=False,
        annotations=True,
        gene_aliases=None,
    ) -> EnrichmentPayload:
        """
        Summarize a list of gene IDs.

        >>> engine = EnrichmentEngine()
        >>> gene_set = GeneSet(gene_ids=...)
        >>> engine.summarize(gene_set)

        :param gene_set: an object representing a list of genes to summarize
        :prompt_template: path to alternative prompt
        :param normalize: whether to normalize labels to IDs
        :param strict: whether to raise an error if a gene symbol cannot be normalized
        :param ontological_synopsis: whether to use the auto-generated synopsis
        :param combined_synopsis: whether to combine the auto-generated and manual synopsis
        :param annotations: whether to use annotations
        :param spike_gene_set: FOR TESTING ONLY
        :return: summary
        """
        if gene_description_source:
            if gene_description_source == GeneDescriptionSource.NONE:
                ontological_synopsis = False
                annotations = False
            elif gene_description_source == GeneDescriptionSource.ONTOLOGICAL_SYNOPSIS:
                ontological_synopsis = True
                annotations = True
            elif gene_description_source == GeneDescriptionSource.NARRATIVE_SYNOPSIS:
                ontological_synopsis = False
                annotations = True
            elif gene_description_source == GeneDescriptionSource.COMBINED_SYNOPSIS:
                combined_synopsis = True
            else:
                raise NotImplementedError
        fill_missing_gene_set_values(gene_set)
        gene_tuples = []
        for gene in gene_set.genes.values():
            logging.info(f"Looking up gene summary for {gene.id}...")
            if combined_synopsis:
                if ontological_synopsis:
                    desc = gene.ontological_synopsis + "; " + gene.narrative_synopsis
                    logging.debug(f"Combined synopsis (priority: auto): {desc}")
                else:
                    desc = gene.narrative_synopsis + "; " + gene.ontological_synopsis
                    logging.debug(f"Combined synopsis (priority: manual): {desc}")
            elif ontological_synopsis:
                desc = gene.ontological_synopsis
                logging.debug(f"Auto synopsis: {desc}")
            else:
                desc = gene.narrative_synopsis
                logging.debug(f"Manual synopsis: {desc}")
            gene_tuples.append((gene.id, gene.symbol, desc))
        logging.info(f"Found {len(gene_tuples)} gene summaries")
        if gene_aliases:
            # for randomization
            gene_tuples = [(id, gene_aliases.get(sym, sym), desc) for id, sym, desc in gene_tuples]
        if not prompt_template:
            prompt_template = str(f"{DEFAULT_ENRICHMENT_PROMPT}.jinja2")
        prompt, tf = self._prompt_from_template(
            gene_tuples,
            template=prompt_template,
            annotations=annotations,
            taxon=gene_set.taxon,
        )
        response_text = self.client.complete(prompt, max_tokens=self.completion_length)
        response_token_length = len(self.encoding.encode(response_text))
        logging.info(f"Response token length: {response_token_length}")
        payload = EnrichmentPayload(
            prompt=prompt,
            response_text=response_text,
            truncation_factor=tf,
            used_auto_synopsis=ontological_synopsis,
            used_combined_synopsis=combined_synopsis,
            annotations=annotations,
            response_token_length=response_token_length,
            model=self.model,
        )
        self.process_payload(payload)
        return payload

    def _prompt_from_template(
        self,
        genes: List[GENE_TUPLE],
        template: Union[str, Path, Template],
        truncation_factor=1.0,
        taxon: str = None,
        annotations=True,
    ) -> Tuple[str, float]:
        if isinstance(template, Path):
            template = str(template)
        if isinstance(template, str):
            # create a Jinja2 template object
            with open(template) as file:
                template = Template(file.read())
        if not isinstance(template, Template):
            raise ValueError(f"Invalid template: {template}")
        gd_tuples = []
        for _, symbol, desc in genes:
            if desc is None:
                desc = ""
            if truncation_factor < 1.0:
                desc = desc[: int(len(desc) * truncation_factor)] + "..."
            gd_tuples.append((symbol, desc))
        prompt = template.render(
            gene_descriptions=gd_tuples,
            annotations=annotations,
            taxon=taxon,
            SUMMARY_KEYWORD=SUMMARY_KEYWORD,
            MECHANISM_KEYWORD=MECHANISM_KEYWORD,
            ENRICHED_TERMS_KEYWORD=ENRICHED_TERMS_KEYWORD,
            END_MARKER=self.end_marker,
        )
        logging.debug(f"Prompt from template: {prompt}")
        logging.info(f"Prompt [{truncation_factor}] Length: {len(prompt)}")
        # https://help.openai.com/en/articles/4936856-what-are-tokens-and-how-to-count-them
        prompt_length = len(self.encoding.encode(prompt)) + 10
        max_len_total = 4097
        if self.model == MODEL_GPT_4:
            max_len_total = 8193
        max_len = max_len_total - self.completion_length
        logging.info(
            f"Prompt [{truncation_factor}] Toks: {prompt_length} / {max_len} Str={len(prompt)}"
        )
        if prompt_length > max_len:  # TODO: check this
            logging.warning(f"Prompt is too long; toks: {prompt_length} len: {len(prompt)}")
            return self._prompt_from_template(
                genes,
                template,
                truncation_factor=truncation_factor * 0.8,
                taxon=taxon,
                annotations=annotations,
            )
        return prompt, truncation_factor

    def map_labels(self, labels: List[str], strict=False) -> Iterator[ENTITY_ID]:
        """Map labels to IDs."""
        for label in labels:
            if ":" in label and " " not in label:
                yield label
            for adapter in self.label_resolvers.values():
                try:
                    curies = [x.upper() for x in adapter.curies_by_label(label)]
                    if len(curies) != 1:
                        logging.warning(f"Found {len(curies)} curies for label: {label}")
                        if strict:
                            raise ValueError(f"Found {len(curies)} curies for label: {label}")
                    if curies:
                        yield curies[0]
                except ValueError:
                    continue

    def add_resolver(self, resolver: str):
        self.label_resolvers[resolver] = get_adapter(resolver)

    def process_payload(self, payload: EnrichmentPayload) -> EnrichmentPayload:
        """Process the payload."""
        toks = re.split(f"{ENRICHED_TERMS_KEYWORD}:", payload.response_text, flags=re.IGNORECASE)
        if len(toks) > 2:
            logging.warning(
                f"Found more than one {ENRICHED_TERMS_KEYWORD} in response: {payload.response_text}"
            )
            toks[1] = " ".join(toks[1:])
        if len(toks) > 1:
            payload.summary = toks[0]
            rest = toks[1]
        else:
            payload.summary = "COULD NOT PARSE"
            rest = payload.response_text
        if "hypothesis:" in rest.lower():
            # this is a little ad-hoc; turbo is prone to inserting a "hypothesis" at the end
            # even though this is not in the prompt
            toks = re.split("hypothesis:", rest, flags=re.IGNORECASE)
            rest = toks[0]
            payload.summary += "\nHypothesis: " + toks[1]
        # we ask to split on ";" but sometimes the model disobeys and uses "," instead
        tok_chars = [";", ",", "-"]
        tokenizations = {}
        for tok_char in tok_chars:
            toks = rest.split(f"{tok_char} ")
            tokenizations[tok_char] = toks
        tokenizations = sorted(tokenizations.items(), key=lambda x: len(x[1]), reverse=True)
        best_tokens = tokenizations[0][1]
        payload.term_strings = [s.lower().strip(":-*;. -\n") for s in best_tokens]  # noqa
        if payload.term_strings and payload.term_strings[-1].startswith("and "):
            # sometimes the LLM will write an oxford comma style list
            payload.term_strings[-1] = payload.term_strings[-1].replace("and ", "")
        payload.term_ids = []
        for term in payload.term_strings:
            payload.term_ids.append(self.normalize_named_entity(term, GeneDescriptionTerm.__name__))
        return payload
